# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2013Jun18

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """A Bloch simulator designed to work with the Spyn-node and optional
    RFwaveforms-node input.  Bloch nodes can be cascaded to effectively build
    segments of a pulse sequence.

    The data are passed from the output of either the Spyn Module (the beginning) or
    the output of a Bloch module to the left output (M_in) of the next Bloch module

    INPUTS:
    M_in - spin data
    G_in - input gradient waveform - if 1 dimensional, this is used for all gradients
           if 2 dimensional, with the 2nd dimension 3, this is used for separate Gx, Gy, Gz waveforms 
           Waveform is linearly interpolated to match the duration of the Bloch module
           if not connected, gradients are all considered constant for Duration
    RF_in - input RF waveform - if 1 dimensional, this is used for RF magnitude
           if 2 dimensional, with the 2nd dimension 2, this is used as specified by the RF Waveform widget
           Waveform is linearly interpolated to match the duration of the Bloch module
           if not connected, rf is considered constant for Duration
    Pars_in - accepts parameters to set other widgets (i.e. if external waveform is used

    OUTPUTS - spin data

    WIDGETS:
    Duration: duration of this particular part of the sequence
    Gx, Gy, Gz - amplitude of gradients.  If an external waveform is supplied, 
                 it is multiplied by these respective values
    RF Mag (uT) - amplitude of RF pulse. If an external waveform is supplied, 
                 it is multiplied by this value.  If RF Flip is set, RF Mag is adjusted to
                 give that flip at the center frequency.
    RF Flip (deg) - flip angle of RF pulse at the center frequency. 
                 If RF Mag is set, RF Flip is adjusted accordingly.
    RF Phase (deg) - phase of RF pulse
    Crusher Tc (ms) - this is used to simulate a crusher, although not too realistically.
                 The (Mx,My) component of all spins decay with an additional exp(-t/Tc) during
                 this module.  If Tc=0, Crusher is ignored.
    RF Waveform: - specifies the mode of the input 2D RF waveform
    """

    def initUI(self):
        # Widgets
        self.addWidget('DoubleSpinBox', 'Duration (ms)', min=0.01, val=1.,decimals=3)
        self.addWidget('DoubleSpinBox', 'Gx (mT/m)', val=0.)
        self.addWidget('DoubleSpinBox', 'Gy (mT/m)', val=0.)
        self.addWidget('DoubleSpinBox', 'Gz (mT/m)', val=0.,decimals=3)
        self.addWidget('DoubleSpinBox', 'RF Mag (uT)', val=0.,decimals=2)
        self.addWidget('DoubleSpinBox', 'RF Flip (deg)', val=0.)
        self.addWidget('DoubleSpinBox', 'RF Phase (deg)', val=0.)
        self.addWidget('DoubleSpinBox', 'Crusher Tc (ms)', val=0.)
        self.addWidget('ExclusivePushButtons','RF Waveform',
                       buttons=['M P(deg)','M P(rad)','M Fq(kHz)','RFx RFy'],val=3)

        # IO Ports
        self.addInPort('M_in', 'NPYarray',obligation=gpi.REQUIRED)
        self.addInPort('G_in', 'NPYarray',obligation=gpi.OPTIONAL)
        self.addInPort('RF_in', 'NPYarray',obligation=gpi.OPTIONAL)
        self.addInPort('Pars_in', 'DICT',obligation=gpi.OPTIONAL)

        self.addOutPort('M_out', 'NPYarray')

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''

        # New data
        m_in = self.getData('M_in')
        rf_in = self.getData('RF_in')

        # read Parameter from RFwaveforms module if changed and available
        if 'Pars_in' in self.portEvents():
            param_dict = self.getData('Pars_in')
            if param_dict is not None:
                if 'Duration (ms)' in param_dict:
                    self.setAttr('Duration (ms)', val=param_dict['Duration (ms)'])
                if 'Gz (mT/m)' in param_dict:
                    self.setAttr('Gz (mT/m)', val=param_dict['Gz (mT/m)'])
                if 'RF Mag (uT)' in param_dict:
                    self.setAttr('RF Mag (uT)', val=param_dict['RF Mag (uT)'])
                if 'RF Flip (deg)' in param_dict:
                    self.setAttr('RF Flip (deg)', val=param_dict['RF Flip (deg)'])
                if 'RF waveform has FM shape' in param_dict:
                    if param_dict['RF waveform has FM shape']:
                        self.setAttr('RF Waveform', val=2)

        # Reconcile timing
        dtms = m_in[10,0,0,0,0,0,0,0,0,0,0]
        durval = max(1.,round(self.getVal('Duration (ms)')/dtms,0))*dtms
        self.setAttr('Duration (ms)',val=durval)

        # Find Area of rf pulse
        if rf_in is not None:
          if rf_in.ndim == 1:
            rf_ave = np.average(rf_in)
            self.setAttr('RF Waveform',visible=False)
          elif rf_in.ndim == 2:
            rf_ave = np.linalg.norm(np.average(rf_in,axis=0))
            self.setAttr('RF Waveform',visible=True)
          else:
            rf_ave = 1000000000 # basically I should make this return an error, don't know how yet
        else:
          rf_ave = 1.;
          self.setAttr('RF Waveform',visible=False)

        # Reconcile flip angle and rfmag
        gamma = m_in[11,0,0,0,0,0,0,0,0,0,0] 

        if 'RF Mag (uT)' in self.widgetEvents():
          rf_mag = self.getVal('RF Mag (uT)')
          rf_flip = 360.*gamma*rf_ave*durval*rf_mag/1000.
          self.setAttr('RF Flip (deg)',val=rf_flip)
        else:
          rf_flip = self.getVal('RF Flip (deg)')
          rf_mag = rf_flip/(360.*gamma*rf_ave*durval)*1000.
          self.setAttr('RF Mag (uT)',val=rf_mag)

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''

        # New data
        m_in = self.getData('M_in')
        g_in = self.getData('G_in')
        rf_in = self.getData('RF_in')

#Constants, per pixel
        m0 = m_in[0,0,:,:,:,:,:,:,:,:,:]
        r1 = m_in[1,0,:,:,:,:,:,:,:,:,:]
        r2 = m_in[2,0,:,:,:,:,:,:,:,:,:]
        vx = m_in[3,0,:,:,:,:,:,:,:,:,:]
        vy = m_in[4,0,:,:,:,:,:,:,:,:,:]
        vz = m_in[5,0,:,:,:,:,:,:,:,:,:]
        fq = m_in[6,0,:,:,:,:,:,:,:,:,:]
        dtms = m_in[10,0,0,0,0,0,0,0,0,0,0]
        gamma = m_in[11,0,0,0,0,0,0,0,0,0,0] 

#Starting values for this module are ending values from last module
        mx = m_in[0,-1,:,:,:,:,:,:,:,:,:]
        my = m_in[1,-1,:,:,:,:,:,:,:,:,:]
        mz = m_in[2,-1,:,:,:,:,:,:,:,:,:]
        x0 = m_in[3,-1,:,:,:,:,:,:,:,:,:]
        y0 = m_in[4,-1,:,:,:,:,:,:,:,:,:]
        z0 = m_in[5,-1,:,:,:,:,:,:,:,:,:]
        t0 = m_in[6,-1,:,:,:,:,:,:,:,:,:]

        t_bloch = self.getVal('Duration (ms)')

        # Some Initial Calcs
        odim = int(t_bloch/dtms)
        oshape = list(m_in.shape)
        oshape[1]=odim
        idim = int(1000.*dtms)
        newdat=np.zeros(tuple(oshape))
        n_iter = int(1000.*t_bloch)
        usec = 0.001 # in units of msec
        radspermT = 2.*np.pi*gamma*usec

        # Synthetic Crusher
        tc = self.getVal('Crusher Tc (ms)')
        if tc > 0:
          r2c = r2 + (1./tc)
        else:
          r2c = r2

        expt1=np.exp(-r1*usec)
        expt2=np.exp(-r2c*usec)
        mz0 = m0*(1.-expt1)
        db0 = fq/gamma

        ######################
        # Create Gx,Gy,Gz array
        ######################
        if g_in is not None:
          gdim = list(g_in.shape)
          x_in  = np.linspace(0.,1.,num=gdim[0])
          x_out = np.linspace(0.,1.,num=n_iter)
          if g_in.ndim==1:
            g_out = np.interp(x_out,x_in,g_in)
            gx = self.getVal('Gx (mT/m)')*g_out
            gy = self.getVal('Gy (mT/m)')*g_out
            gz = self.getVal('Gz (mT/m)')*g_out
          elif g_in.ndim==2 and list(g_in.shape)[1]==3:
            gx = self.getVal('Gx (mT/m)')*np.interp(x_out,x_in,g_in[:,0])
            gy = self.getVal('Gy (mT/m)')*np.interp(x_out,x_in,g_in[:,1])
            gz = self.getVal('Gz (mT/m)')*np.interp(x_out,x_in,g_in[:,2])
          else:
            g_out = np.ones(n_iter)
            gx = self.getVal('Gx (mT/m)')*g_out
            gy = self.getVal('Gy (mT/m)')*g_out
            gz = self.getVal('Gz (mT/m)')*g_out
        else:
          g_out = np.ones(n_iter)
          gx = self.getVal('Gx (mT/m)')*g_out
          gy = self.getVal('Gy (mT/m)')*g_out
          gz = self.getVal('Gz (mT/m)')*g_out

        ######################
        # Create RFx, RFy arrays
        ######################
        crfph = np.cos(np.radians(self.getVal('RF Phase (deg)')))
        srfph = np.sin(np.radians(self.getVal('RF Phase (deg)')))
        rfm = 0.001 * self.getVal('RF Mag (uT)')
        if rf_in is not None:
          rfdim = list(rf_in.shape)
          x_in  = np.linspace(0.,1.,num=rfdim[0])
          x_out = np.linspace(0.,1.,num=n_iter)
          if rf_in.ndim==1:
            rf_mag = np.interp(x_out,x_in,rf_in)
            rfx = rfm*crfph*rf_mag
            rfy = rfm*srfph*rf_mag
          elif rf_in.ndim==2 and list(rf_in.shape)[1]==2:
            rfwave = self.getVal('RF Waveform')
            rf0 = np.interp(x_out,x_in,rf_in[:,0])
            rf1 = np.interp(x_out,x_in,rf_in[:,1])
            if rfwave == 0: # M P(deg)
              rf_tx = rf0*np.cos(np.radians(rf1))
              rf_ty = rf0*np.sin(np.radians(rf1))
            if rfwave == 1: # M P(rad)
              rf_tx = rf0*np.cos(rf1)
              rf_ty = rf0*np.sin(rf1)
            if rfwave == 2: # M F(kHz)
              rfph = 2.*np.pi*0.001*np.cumsum(rf1) # phase is integral of frequency, and dt is 0.001 ms
              rf_tx = rf0*np.cos(rfph)
              rf_ty = rf0*np.sin(rfph)
            if rfwave == 3: # RFx RFy
              rf_tx = rf0
              rf_ty = rf1
            rfx = rfm*(crfph*rf_tx - srfph*rf_ty)
            rfy = rfm*(crfph*rf_ty + srfph*rf_tx)
          else:
            rf_out = np.ones(n_iter)
            rfx = crfph*self.getVal('RF Mag (uT)')/1000.*rf_out
            rfy = srfph*self.getVal('RF Mag (uT)')/1000.*rf_out
        else:
          rf_out = np.ones(n_iter)
          rfx = crfph*self.getVal('RF Mag (uT)')/1000.*rf_out
          rfy = srfph*self.getVal('RF Mag (uT)')/1000.*rf_out

        ######################
        # Now iterate through time
        ######################
        # record data in newdat for each index in outer loop
        # inner loop interval is 1 usec
        ######################

        for outer in range(0,odim): 
          for inner in range(0,idim): 
            i = outer*idim+inner
            time = float(i+1)*usec
            x = x0 + time*vx
            y = y0 + time*vy
            z = z0 + time*vz

            xph = radspermT*rfx[i] #radians to rotate about x
            yph = radspermT*rfy[i] #radians to rotate about y
            zph = radspermT*(x*gx[i]+y*gy[i]+z*gz[i] + db0) #radians to rotate about z

            mx,my,mz = self.rotate(mx,my,mz,xph,yph,zph)
            mx = mx*expt2
            my = my*expt2
            mz = mz0 + expt1*mz

          newdat[0,outer,:,:,:,:,:,:,:,:,:] = mx
          newdat[1,outer,:,:,:,:,:,:,:,:,:] = my
          newdat[2,outer,:,:,:,:,:,:,:,:,:] = mz
          newdat[3,outer,:,:,:,:,:,:,:,:,:] = x
          newdat[4,outer,:,:,:,:,:,:,:,:,:] = y
          newdat[5,outer,:,:,:,:,:,:,:,:,:] = z
          newdat[6,outer,:,:,:,:,:,:,:,:,:] = t0+time
          newdat[7,outer,:,:,:,:,:,:,:,:,:] = gx[i]
          newdat[8,outer,:,:,:,:,:,:,:,:,:] = gy[i]
          newdat[9,outer,:,:,:,:,:,:,:,:,:] = gz[i]
          newdat[10,outer,:,:,:,:,:,:,:,:,:] = rfx[i]
          newdat[11,outer,:,:,:,:,:,:,:,:,:] = rfy[i]
          newdat[12,outer,:,:,:,:,:,:,:,:,:] = tc

        m_out=np.concatenate((m_in,newdat),axis=1)

        self.setData('M_out', m_out)

        return 0

    ######################
    def rotate(self, rx, ry, rz, px, py, pz):
    ######################
# R = (rx, ry, rz) = (mx, my, mz)
# P = (px, py, pz)
# Rotate R about P, by |P| radians
#
# R1 is the projection of R on P, which does not rotate
# R2 is the residue of R which is perpendicular to P
# R3 and R4 are the vectors of the rotated R2 which are
#           parallel and perpendicular to R2, respectively
#
# R1 = P (P dot R)/(|P|**2)
# R2 = R - R1
# R3 = R2 Cos(|P|)
# R4 = (P x R2) Sin(|P|) / |P|
#
# R' = R1+R3+R4

      psqu = px*px + py*py + pz*pz
      if (psqu > 0).any():
        pmag = np.sqrt(psqu)

        # Need to mask, to avoid div_by_0
        pzero = np.zeros(psqu.shape)

        pdotr = px*rx + py*ry + pz*rz
        r1norm = np.where(psqu,pdotr/psqu,pzero)
        cp = np.cos(pmag)
        spn = np.where(pmag,np.sin(pmag)/pmag,pzero)

        r1x = px*r1norm
        r1y = py*r1norm
        r1z = pz*r1norm

        r2x = rx-r1x
        r2y = ry-r1y
        r2z = rz-r1z

        r3x = r2x*cp
        r3y = r2y*cp
        r3z = r2z*cp

        r4x = (py*r2z - pz*r2y)*spn
        r4y = (pz*r2x - px*r2z)*spn
        r4z = (px*r2y - py*r2x)*spn

        rx = r1x+r3x+r4x
        ry = r1y+r3y+r4y
        rz = r1z+r3z+r4z

      return rx,ry,rz

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Zhiqiang Li 
# Date: 2015nov20
# Note: Modified from Dimensions witten by Ryan Robison.

import gpi
import numpy as np

#################################
#################################

class ExternalNode(gpi.NodeAPI):
    """A module for splitting a specified dimension of the input data.
       InData: The data to be splitted.
       Split Dimension: The specific dimension to be splitted.
       Fix: User specifies either new low or high dimension size.
       Fixed Dim Size: the "new size" of the fixed low or high dimension.
    """

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'Info:')
        self.maxdim = 8
        self.ndim = self.maxdim
        dim_buttons = []
        for i in range(-self.maxdim, 0):
           dim_buttons.append(str(i))
        self.addWidget('ExclusivePushButtons', 'Split Dimension',
                buttons=dim_buttons, val=-1)
        self.addWidget('ExclusivePushButtons', 'Fix',
                buttons=['Low Dim', 'High Dim'],val=0)
        self.addWidget('Slider', 'Fixed Dim Size', val=1, min=1, max=32)
        self.addWidget('PushButton', 'Compute', toggle=True, val=0)

        # IO Ports
        self.addInPort('InData', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('outData', 'NPYarray')

    def validate(self):
        self.log.info("Start Validation")
        data = self.getData('InData')
        if data is not None:
            self.log.info("    Validate selected dimenions")
            sel = self.getVal('Split Dimension')
            if sel >= 0:
                sel = sel - self.ndim
            if (-sel) > data.ndim:
                self.log.warn("Try to split a dimension beyond " + \
                        "the input dimensions!")
                return (1)

            dim_buttons = []
            for i in range(-data.ndim, 0):
                dim_buttons.append(str(i))
            self.setAttr('Split Dimension', buttons=dim_buttons, val=sel)
            self.maxdim = data.ndim
            self.ndim = self.maxdim

            self.log.info("    Validate desired new size")
            fixedDimSize = self.getVal("Fixed Dim Size")
            size0 = int(data.shape[sel]/fixedDimSize)*fixedDimSize
            if data.shape[sel] != size0:
                self.log.warn("Can't split " + str(int(data.shape[sel])) + \
                        " to " + str(fixedDimSize))
                return (1)
            sel = self.getVal('Split Dimension')
            if sel >= 0:
                sel = sel - self.ndim
            LowHighFlag = self.getVal("Fix")
            if LowHighFlag == 0: # set low dimension to new size
                lowDimSize = fixedDimSize
                highDimSize = data.shape[sel]/lowDimSize
            else:
                highDimSize = fixedDimSize
                lowDimSize = data.shape[sel]/highDimSize
            newDim = list(data.shape[-data.ndim:sel]) + [int(highDimSize)] + \
                    [int(lowDimSize)]
            if sel < -1:
                newDim = newDim + list(data.shape[(sel+1):])

            info = "Input:  " + str(data.shape) + "\n" \
                   "Output: " + str(newDim)
            self.setAttr("Info:", val=info)
        return (0)
         
    def compute(self):

        if self.getVal('Compute') is True:
            data = self.getData('InData')
            sel = self.getVal('Split Dimension')
            if sel >= 0:
                sel = sel - self.ndim
            fixedDimSize = self.getVal("Fixed Dim Size")
            LowHighFlag = self.getVal("Fix")
            if LowHighFlag == 0: # set low dimension to new size
                lowDimSize = fixedDimSize
                highDimSize = data.shape[sel]/lowDimSize
            else:
                highDimSize = fixedDimSize
                lowDimSize = data.shape[sel]/highDimSize
            newDim = list(data.shape[-data.ndim:sel]) + [int(highDimSize)] + [int(lowDimSize)]
            if sel < -1:
                newDim = newDim + list(data.shape[(sel+1):])
            splittedData = data.reshape(newDim)
            self.setData("outData", splittedData)
        return(0)

    def execType(self):
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012dec10

import gpi
import numpy as np


class ExternalNode(gpi.NodeAPI):
    """A node that allows cyclic connections and provides a countdown for
    iterative algorithms using the conjugate gradient approach.  The code tries to
    match the nomenclature of the excellent article by Jonathan Richard Shewchuk,
    An Introduction to the Conjugate Gradient Methods Without the Agonizing Pain
    and in particular equations 45-49 on page 32 of that manuscript.

    The node will automatically step through a specified number of iterations

    So we are solving for Ax = b
    INPUTS: The first four inputs are initialization, the last one is for iteration
    d_in - initial guess for direction to move, equals b-Ax (Eq. 45)
    r_in - initial guess, same as d_in, equals b-Ax (Eq. 45)
    x_in - initial guess for x
    Ad_in - matrix A times d_in
    Adloop_in - used during iterations, this takes A multiplied by d_out (below)

    OUTPUTS:
    d_out - multiply this vector by A and send back to Adloop_in
    r_out - not really used, just FYI
    x_out - the final answer

    WIDGETS:
    Current Iteration - lets you know which iteration is being performed
    Max Iteration - enter the desired number of iterations
    Start/Stop - click on to start, click off to stop before iterations are done
    Step - click to step through one more iteration
    Reset - click to reset iteration count and begin with initial conditions
    """

    def initUI(self):
        # Widgets
        self.addWidget(wdg='TextBox', title='Current Iteration', val='0')
        self.addWidget(wdg='SpinBox', title='Max Iteration', min=0)
        self.addWidget(wdg='PushButton', title='Start/Stop', toggle=True)
        self.addWidget(wdg='PushButton', title='Step')
        self.addWidget(wdg='PushButton', title='Reset')

        # IO Ports
        self.addInPort(title='d_in', type='NPYarray')
        self.addInPort(title='r_in', type='NPYarray')
        self.addInPort(title='x_in', type='NPYarray')
        self.addInPort(title='Ad_in', type='NPYarray')
        self.addInPort(title='Adloop_in', type='NPYarray', obligation=gpi.OPTIONAL, cyclic=True)

        self.addOutPort(title='d_out', type='NPYarray')
        self.addOutPort(title='r_out', type='NPYarray')
        self.addOutPort(title='x_out', type='NPYarray')

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''

        # if init changes then reset the count and stop iteration
        if set(self.portEvents()).intersection(set(['d_in', 'r_in', 'x_in', 'Ad_in'])):
            self.setAttr('Reset', val=True)

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''
        reset = self.getVal('Reset')
        on = self.getVal('Start/Stop')
        iternum = int(self.getVal('Current Iteration'))
        maxiter = self.getVal('Max Iteration')
        step = self.getVal('Step')

        if step:
            maxiter += 1
            self.setAttr('Max Iteration', val=maxiter)
            on = True
            self.setAttr('Start/Stop', val=on)

        if reset:
            self.setAttr('Start/Stop', val=False)
            self.setAttr('Current Iteration', val='0')
            d_in = self.getData('d_in')
            r_in = self.getData('r_in')
            x_in = self.getData('x_in')
            self.setData('d_out', d_in)
            self.setData('r_out', r_in)
            self.setData('x_out', x_in)

        elif on:
            if iternum >= maxiter:
                self.setAttr('Start/Stop', val=False)
                return 0
            elif iternum == 0:
                self.setAttr('Current Iteration', val=str(iternum+1))

                d_in = self.getData('d_in')
                r_in = self.getData('r_in')
                x_in = self.getData('x_in')
                Ad_in = self.getData('Ad_in')

                (d_out, r_out, x_out) = self.do_cg(d_in, r_in, x_in, Ad_in)

                self.setData('d_out', d_out)
                self.setData('r_out', r_out)
                self.setData('x_out', x_out)

            else:
                self.setAttr('Current Iteration', val=str(iternum+1))

                Adloop_in = self.getData('Adloop_in')

                # copy last iter
                d_in = self.getData('d_out')
                r_in = self.getData('r_out')
                x_in = self.getData('x_out')

                (d_out, r_out, x_out) = self.do_cg(d_in, r_in, x_in, Adloop_in)

                self.setData('d_out', d_out)
                self.setData('r_out', r_out)
                self.setData('x_out', x_out)
        else:
            self.setData('d_out', None)
            self.setData('r_out', None)
            self.setData('x_out', None)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_THREAD


    def do_cg(self, d_in, r_in, x_in, Ad_in):

        d_out = np.zeros_like(d_in)
        r_out = np.zeros_like(r_in)
        x_out = np.zeros_like(x_in)

        # Calculate alpha
        # r^H r / (d^H Ad)
        rHr = np.dot(np.conj(r_in.flatten()), r_in.flatten()) 
        dHAd = np.dot(np.conj(d_in.flatten()), Ad_in.flatten())
        alpha = rHr/dHAd

        # Calculate x(i+1)
        # x(i) + alpha d(i)
        x_out = x_in + alpha * d_in

        # Calculate r(i+1)
        # r(i) - alpha Ad(i)
        r_out = r_in - alpha * Ad_in

        # Calculate beta
        # r(i+1)^H r(i+1) / (r(i)^H r(i))
        r1Hr1 = np.dot(np.conj(r_out.flatten()), r_out.flatten()) 
        beta = r1Hr1 / rHr

        # Calculate d(i+1)
        # r(i+1) + beta d(i)
        d_out = r_out + beta * d_in

        return (d_out, r_out, x_out)




# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2014Jan

import gpi


class ExternalNode(gpi.NodeAPI):

    """Node to read either real and imaginary or magnitude and phase data and convert to complex data

    INPUT:
    inLeft: input numpy array
    inRight: input numpy array

    OUTPUT:
    out:  output numpy array

    WIDGETS:
    Operation:  Determines what is done
      If both input ports (L and R) are populated
        L+iR - combines the two fields of same size to complex
          data inLeft - Real
          data inRight - Imaginary
        L exp(iR) - combines the two fields of same size to complex
          data inLeft - Magnitude
          data inRight - Phase
      If a single input port P is populated
        Real - output complex data with input's real channel and all zeros in the imaginary channel
        Imag - output complex data with input's imaginary channel and all zeros in the real channel
        Phase - output complex data with magnitude = 1 and phase given by input
        Vec2Cmplx - if the last dimension is 2, change this to R/I

     Phase: pick degrees or radians when appropriate

     OutType: single or double precision complex type
    """

    def initUI(self):
        self.addInPort('inLeft', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addInPort('inRight', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray')

        self.addWidget('ExclusiveRadioButtons', 'Operation', buttons=['Real', 'Imag', 'Phase', 'Vec2Cmplx'], val=0)
        self.addWidget('ExclusiveRadioButtons', 'Combine Operation', buttons=['L+iR', 'L exp(iR)'], val=0)

        self.addWidget('ExclusivePushButtons', 'Phase', buttons=['Deg', 'Rad', 'Cyc'], val=0)
        self.addWidget('ExclusivePushButtons', 'OutType', buttons=['Complex64', 'Complex128'], val=0)

    def validate(self):
        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')
        singleCase = 0

        # No input ports connected
        if (data1 is None and data2 is None):
            self.setAttr('Operation', visible=False)
            self.setAttr('Combine Operation', visible=False)

        # Both input ports connected
        elif (data1 is not None and data2 is not None):
            if data1.shape != data2.shape:
                self.log.warn("data shapes must match")
                return(1)
            else:
                self.setAttr('Combine Operation', visible=True)
                self.setAttr('Operation', visible=False)

            if self.getVal('Combine Operation') == 1:
                self.setAttr('Phase', visible = True)
            else:
                self.setAttr('Phase', visible = False)

        # Left input port connected
        else:
            if data1 is not None:
                if data1.shape[-1] == 2:
                    singleCase = 1
            else:
                if data2.shape[-1] == 2:
                    singleCase = 2

            if singleCase != 0:
                self.setAttr('Operation', visible=True)
                self.setAttr('Combine Operation', visible=False)
                self.setAttr('Operation', buttons=['Real', 'Imag', 'Phase', 'Vec2Cmplx'])
            else:
                self.setAttr('Operation', visible=True)
                self.setAttr('Combine Operation', visible=False)
                self.setAttr('Operation', buttons=['Real', 'Imag', 'Phase'])

            if self.getVal('Operation') == 2:
                self.setAttr('Phase', visible = True)
            else:
                self.setAttr('Phase', visible = False)

        return(0)

    def compute(self):
        import numpy as np

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')
        op = self.getVal('Operation')
        cop = self.getVal('Combine Operation')
        phase = self.getVal('Phase')
        outtype = self.getVal('OutType')

        # Both ports connected: 'L+iR' or 'L exp(iR)'
        if (data1 is not None) and (data2 is not None):
            if cop == 0:  # L+iR
                out = data1 + 1j * data2
            elif cop == 1:  # L exp(iR)
                if phase == 0:  # Degrees
                    out = data1 * np.exp(1j * np.radians(data2))
                elif phase == 1: # radians
                    out = data1 * np.exp(1j * data2)
                else: # cycles
                    out = data1 * np.exp(1j * 2 * np.pi * data2)
            else:
                out = None

        # Left port connected
        elif (data1 is not None or data2 is not None):
            if data1 is None: # Swap if data1 is emtpy
                data1 = data2

            if op == 0:   # Real
                out = data1.astype(complex)
            elif op == 1:  # Imag
                out = data1 * 1j
            elif op == 2:  # Phase
                if phase == 0:  # Degrees
                    out = np.exp(1j * np.radians(data1))
                elif phase == 1: # Radians
                    out = np.exp(1j * data1)
                else: # cycles
                    out = np.exp(1j * 2 * np.pi * data1)
            elif op == 3:  # Vec2Cmplx
                if data1.shape[-1] != 2:
                    self.log.warn('The \'Vec2Cmplx\' option is not valid for vec != 2')
                    return 1
                out = data1[..., 0] + 1j * data1[..., 1]
            else:
                out = None

        # No port connected
        else:
            out = None

        # Output data if at least one port connected
        if out is not None:
            if outtype == 0:
                self.setData('out', out.astype(np.complex64))
            elif outtype == 1:
                self.setData('out', out.astype(np.complex128))

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


#__author__ = "Nick Zwart"
#__date__ = "2010dec07"
''' For converting between vectorized matrices and complex matrices
    vectorized:
        x[m,n,1]   is a 2D 1-vec array
        x[m,n,o,2] is a 3D 2-vec array
    complex:
        x[0,0] = 1+j2 is a complex element from a 2D array

'''

import numpy as np  # operate on numpy arrays

# newer, better, faster


def c_2_vec(in1, typ):
    '''Convert a complex (or real valued) array to a 2-vec
    '''

    # reshape
    new_shape = list(in1.shape)
    new_shape += [2]  # add vector

    # allocate new 2-vec mem
    out = np.zeros(new_shape, typ)

    out[(Ellipsis, 0)] = in1.real
    out[(Ellipsis, 1)] = in1.imag

    return(out)


def vec_2_c(in1, typ):
    '''Convert a vectorized 1 or 2-vec to complex.
       -Unfortunately this requires a math operation
        instead of just recasting (this should change
        in the future).
    '''

    # get input array shape
    new_shape = list(in1.shape)
    veclen = new_shape[-1]

    # check that last dim is either 1 or 2,
    # if not, then skip reshape
    if(veclen == 1 or veclen == 2):
        new_shape.pop()  # remove vec dim

    # allocate new complex data type
    out = np.zeros(new_shape, typ)

    # make 2-vec complex
    if veclen == 1:
        out[:] = in1[:, 0]
    elif veclen == 2:
        out[:] = in1[(Ellipsis, 0)] + in1[(Ellipsis, 1)]*1j
    else:  # real valued
        out[:] = in1[:]

    return(out)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2015nov13

import numpy as np
import gpi


class NonExclusivePushButtons2(gpi.NonExclusivePushButtons):
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

    # overwrite set_val
    def set_val(self, value):
        if type(value) is not list:
            value = [value]

        # remove any values that don't have a corresponding button
        names = [button.text() for button in self.buttons]
        temp = []
        for i in range(len(value)):
            if value[i] in names:
                temp.append(value[i])
        value = temp

        self._value = value
        for button_ind in range(len(self.buttons)):
            name = self.buttons[button_ind].text()
            if name in value:
                self.buttons[button_ind].setChecked(True)
            else:
                self.buttons[button_ind].setChecked(False)

    # overwrite findValue
    def findValue(self, value):
        valarr = []
        for button in self.buttons:
            if button.isChecked():
                valarr.append(button.text())
        self._value = valarr
        return


class ExternalNode(gpi.NodeAPI):
    """Flips/mirrors data along specified dimension(s)
    INPUT - input array
    OUTPUT - output array

    WIDGETS:
    """

    def initUI(self):

        # Widgets
        self.ndim = 9  # 9 for now
        self.button_labels = [str(dim) for dim in (list(range(-9,0)))]
        # self.button_labels = ['-2', '-1']
        self.addWidget('NonExclusivePushButtons2', 'Flip Dimensions:',
                       buttons=self.button_labels, val=[])

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''update the widgets based on the input arrays
        '''

        data = self.getData('in')
        fdims = self.getVal('Flip Dimensions:')

        # find max selected dim
        if self.portEvents() and len(fdims) > 0:
            max_fdim = max([-int(dim) for dim in fdims])
            if max_fdim > data.ndim:
                self.log.warn('Data dimensionality has changed! The selected \
flip dimensions are not supported by current data \
dimensionality.')
                return(1)

        # update dim button labels, remove buttons if only 1D
        dim_buttons = list(np.arange(-data.ndim, 0).astype('str'))
        if data.ndim > 1:
            self.setAttr('Flip Dimensions:', buttons=dim_buttons, val=fdims,
                         visible=True)
        else:
            fdims = ['-1']
            self.setAttr('Flip Dimensions:', val=fdims, visible=False)

        return(0)

    def compute(self):

        data = self.getData('in')
        fdims = self.getVal('Flip Dimensions:')

        flip_string = 'out = data['
        for i in range(-data.ndim, 0):
            if str(i) in fdims:
                flip_string = flip_string+'::-1, '
            else:
                flip_string = flip_string+':, '
        flip_string = flip_string[:-2]+']'

        g = globals()
        l = locals()
        exec(flip_string, g, l)
        out = l['out']
        self.setData('out', out)

        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


#Author: Nick Zwart
#Date:  2012sep02
#Payal Bhavsar added read for ASCII type
#Date: Nov2013

import os
import sys
import time
import numpy as np
import gpi


def read(fname, datatype, numelem, skipbytes):
    ''' Read and cast data arrays from raw file format.
    -datatype: 0:float, 1:double, 2:int, 3:char, 4:complex<float>,
    5:complex<double>
    '''
    if not type(fname) is str:
        self.node.warn("Input is not a string.")
        raise Exception("read(): input is not a string.")

    # opens the file
    try:
        fil = open(fname, 'rb')
    except IOError:
        print('cannot open', fname)
        raise Exception("read(): cannot open file")

    # read into byte array
    fil.seek(skipbytes)

    # read the dataset 
    if datatype == 0:
        data = fil.read(numelem*np.dtype(np.float32).itemsize)
        data = np.fromstring(data, dtype=np.float32)
    elif datatype == 1:
        data = fil.read(numelem*np.dtype(np.float64).itemsize)
        data = np.fromstring(data, dtype=np.float64)
    elif datatype == 2:
        data = fil.read(numelem*np.dtype(np.int32).itemsize)
        data = np.fromstring(data, dtype=np.int32)
    elif datatype == 3:
        data = fil.read(numelem*np.dtype(np.uint8).itemsize)
        data = np.fromstring(data, dtype=np.uint8)
    elif datatype == 4:
        data = fil.read(numelem*np.dtype('i2').itemsize)
        data = np.fromstring(data, dtype='i2')
    elif datatype == 5:
        data = fil.read(numelem*np.dtype(np.complex64).itemsize)
        data = np.fromstring(data, dtype=np.complex64)
    elif datatype == 6:
        data = fil.read(numelem*np.dtype(np.complex128).itemsize)
        data = np.fromstring(data, dtype=np.complex128)
    else:
        print("Error: read(): No valid datatype chosen: "+str(datatype))
        raise Exception("read(): no valid datatype chosen.")

    # convert data
    fil.close()
    return(data)


class ExternalNode(gpi.NodeAPI):
    """Reads raw data into several c-types: float, double, int, and char.
    OUTPUT: Numpy array of data

    WIDGETS:
    I/O Info: information about file
    File Browser: file browswer
    Skip Bytes: number of bytes past which to start reading data (e.g. size of header)
    <type>: type of data in file
    ndim:  Number of dimensions to fill (of output array)
    Dimension N: number of elements in the Nth dimension
    Compute: compute
    """

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'I/O Info:')
        self.addWidget(
            'OpenFileBrowser', 'File Browser', button_title='Browse',
            caption='Open File', filter='raw (*.raw);;all (*)')
        self.addWidget('SpinBox', 'Skip Bytes:', val=0, min=0)
        self.addWidget('ExclusiveRadioButtons', '<type>', buttons=[
                       'float', 'double', 'int', 'char','short int', 'complex<float>',
                       'complex<double>'], val=0)

        self.ndim = 8
        self.dim_base_name = 'Dimension: '
        self.addWidget('Slider', 'ndim:', max=self.ndim, val=1, min=1)
        for i in range(self.ndim):
            self.addWidget('SpinBox', self.dim_base_name+str(i), min=1, val=1, max=gpi.GPI_INT_MAX)

        self.addWidget('PushButton', 'Compute', toggle=True)

        # IO Ports
        self.addOutPort('out', 'NPYarray')

        # store for later use
        self.URI = gpi.TranslateFileURI

    def validate(self):
        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

    def compute(self):

        import os
        import sys
        import time
        import numpy as np

        # visibility
        cdim = self.getVal('ndim:')
        dims = []
        nelem = 1
        for i in range(self.ndim):
            if i < cdim:
                self.setAttr(self.dim_base_name+str(i), visible=True)
                dims.append(self.getVal(self.dim_base_name+str(i)))
                nelem = nelem*self.getVal(self.dim_base_name+str(i))
            else:
                self.setAttr(self.dim_base_name+str(i), visible=False)

        # start file browser
        fname = self.URI(self.getVal('File Browser'))

        # check that the path actually exists
        if not os.path.exists(fname):
            self.log.node("Path does not exist: "+str(fname))
            return 0

        # show some file stats
        fstats = os.stat(fname)
        # creation
        ctime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_ctime))
        # mod time
        mtime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_mtime))
        # access time
        atime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_atime))
        # filesize
        fsize = fstats.st_size
        # user id
        uid = fstats.st_uid
        # group id
        gid = fstats.st_gid

        info = "created: "+str(ctime)+"\n" \
               "accessed: "+str(atime)+"\n" \
               "modified: "+str(mtime)+"\n" \
               "UID: "+str(uid)+"\n" \
               "GID: "+str(gid)+"\n" \
               "file size (bytes): "+str(fsize)+"\n"
        self.setAttr('I/O Info:', val=info)

        # compute block
        compute = self.getVal('Compute')
        dtype = self.getVal('<type>')
        skipbytes = self.getVal('Skip Bytes:')

        if not compute:
            return 0

        out = read(fname, dtype, nelem, skipbytes)

        try:
            out.shape = dims
        except:
            self.log.warn("ERROR: ReadRaw(): reshape: dims are not the " \
                + "same footprint as array!")
            return 0

        self.setData('out', out)

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.

# Author: Nick Zwart
# Date: 2014mar21

import gpi
import os
import h5py

class ExternalNode(gpi.NodeAPI):
    """Reads HDF5 files using the h5py project libs.

    NOTE: This is a simple reader for getting 'dataset' objects from an HDF5 file.
    Since the files can be built in many different ways, this should be considered
    a good starting point for writing code to read your specific format.  More
    general use cases will be added in future releases.
    """

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'I/O Info:')
        self.addWidget(
            'OpenFileBrowser', 'File Browser', button_title='Browse',
            caption='Open File', filter='hdf5 (*.hdf5);;hdf5 (*.h5)')
        self.addWidget('ComboBox', 'dataset')
        self.addWidget('PushButton', 'reload datasets')
        self.addWidget('PushButton', 'read', toggle=True, val=False)
        # self.addWidget('StringBox', 'set-name', val='gpidata')

        # IO Ports
        self.addOutPort('out', 'NPYarray')

        # store for later use
        self.URI = gpi.TranslateFileURI
        self.fname = None

    def validate(self):
        self.fname = self.URI(self.getVal('File Browser'))

        if ('File Browser' in self.widgetEvents()
             or 'reload datasets' in self.widgetEvents()):
            self.setAttr('read', val=False)

            # check that the path actually exists
            if not os.path.exists(self.fname):
                self.log.node("Path does not exist: " + self.fname)
                return 1

            f = h5py.File(self.fname, "r")

            # show available keys
            self.dataset_names = []
            def append_if_dataset(name, obj):
                if isinstance(obj, h5py.Dataset):
                    self.dataset_names.append(name)
            f.visititems(append_if_dataset)
            self.setAttr('dataset', items=self.dataset_names)

            f.close()

        if self.getVal('dataset') is not None:
            self.setDetailLabel(self.fname + "::" + self.getVal('dataset'))
        else:
            self.setDetailLabel(self.fname)

        return 0

    def compute(self):
        import time
        import numpy as np

        # check that the path actually exists
        if self.fname is None or not os.path.exists(self.fname):
            self.log.node("Path does not exist: " + str(self.fname))
            return 0

        # show some file stats
        fstats = os.stat(self.fname)
        # creation
        ctime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_ctime))
        # mod time
        mtime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_mtime))
        # access time
        atime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_atime))
        # filesize
        fsize = fstats.st_size
        # user id
        uid = fstats.st_uid
        # group id
        gid = fstats.st_gid

        info = "created: "+str(ctime)+"\n" \
               "accessed: "+str(atime)+"\n" \
               "modified: "+str(mtime)+"\n" \
               "UID: "+str(uid)+"\n" \
               "GID: "+str(gid)+"\n" \
               "file size (bytes): "+str(fsize)+"\n" \

        self.setAttr('I/O Info:',val=info)

        # get file pointer
        f = h5py.File(self.fname, "r")

        setname = self.getVal('dataset')

        if self.getVal('read'):
            # assuming the dataset is a numpy array
            data = f[setname].value
            self.setData('out', data)

            f.close()

        return 0



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2013 Aug

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """Gridding module for Post-Cartesian Data - works with 2D and 3D data

    INPUTS:
    data - k-space complex data - if not supplied, Grid uses "1" for all of its data
    coords - k-space coordinates, normalized in units of "1/resolution", i.e. ranging from -0.5 to 0.5
               Last dimension must be 1, 2 or 3 (corresponding to kx, kx/ky, or kx/ky/kz, respectively)
               Coordinates at the very edge of gridded k-space then have values of -/+ 0.5
    weighting - sampling density correction - if not supplied, Grid uses "1"
    params_in - optional dictionary from (e.g.) spiralcoords, to automatically specify the effective matrix

    OUTPUTS:
    gridded data, which is M+E dimensions, where
                  M is 1, 2 or 3, depending on last dimension of coords input (i.e. 1D, 2D, or 3D)
                    The size of these M dimensions is 1.5 times the given effective matrix
                  E is (# input data dims) - (Dims per set widget value)
                    E represents slices, coils, etc., which don't get gridded together.

    WIDGETS:
    Dims per Set - How many dimensions get gridded into the same space (and the space is determined by the last dim of coords)
                   Remaining dimensions are independent, e.g. for slices, coils, etc.
    Eff MTX XY - number of pixels in the final image (XY), nominally (without zero-padding) given by FOV/resolution
      Add 25% to matrix for "true resolution" for (e.g.) spiral
      Output data are gridded to a matrix 50% larger than this to mitigate gridding artifacts
    Eff MTX Z - number of pixels in the final image (Z), nominally (without zero-padding) given by FOV/resolution
      Add 25% to matrix for "true resolution" for (e.g.) stack of cones, spherical distributed spiral, FLORET
      Output data are gridded to a matrix 50% larger than this to mitigate gridding artifacts
    dx, dy, dz - for off-center FOV correction.  Specify number of pixels in each direction to shift (in image space)
                 prior to gridding.

      Note on Input dimensions: If coords is N dimensions, with the last used for the 2D/3D information,
                                  1) weights must have N-1 dimensions, of the same shape as corresponding coords
                                  2) data can have N-1 or more dimensions (of correct shape)
                                  3) If data has extra dimension (the first dimensions), data from each index
                                       in these dimensions are gridded using the same coords and weights

      Examples:  2D Spiral Gridding, 4-channel coil, 22 slices, 32 arms, 4056 points per interleaf
                     data are complex, 4 x 22 x 32 x 4056
                     coords are real,      22 x 32 x 4056 x 2
                     weights are real,     22 x 32 x 4056
                     Dims per Set widget is 2
                     Eff Mtx XY widget is 200
                     output data size is 4 x 22 x 300 x 300

                 3D Distributed Spiral Gridding, 8 coils, 320 arms, 4056 points per interleaf
                     data are complex, 8 x 320 x 4056
                     coords are real,      320 x 4056 x 3
                     weights are real,     320 x 4056
                     Dims per Set widget is 2
                     Eff Mtx XY widget is 160
                     Eff Mtx Z widget is 120
                     output data size is 8 x 180 x 240 x 240
    """

    def execType(self):
        return gpi.GPI_PROCESS

    def initUI(self):

        # Widgets
        self.addWidget('Slider','Dims per Set',min=1)
        self.addWidget('SpinBox','Eff MTX XY', min=5, val=240)
        self.addWidget('SpinBox','Eff MTX Z',  min=5, val=240)
        self.addWidget('DoubleSpinBox','dx (pixels)', val=0.0)
        self.addWidget('DoubleSpinBox','dy (pixels)', val=0.0)
        self.addWidget('DoubleSpinBox','dz (pixels)', val=0.0)

        # IO Ports
        self.addInPort('data', 'NPYarray', dtype=[np.complex64, np.complex128],
                       obligation=gpi.OPTIONAL)
        self.addInPort('coords', 'NPYarray', dtype=[np.float64, np.float32],
                       obligation=gpi.REQUIRED)
        self.addInPort('weighting', 'NPYarray', dtype=[np.float32, np.float64],
                       obligation=gpi.OPTIONAL)
        self.addInPort('params_in', 'DICT', obligation = gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray', dtype=[np.complex64, np.complex128])

    def validate(self):
        ''' DHW check if input data, coords and weight have the correct dimentions and sizes.
        '''

        data = self.getData('data')
        crds = self.getData('coords')
        wghts = self.getData('weighting')
        inparam = self.getData('params_in')

        if crds.shape[crds.ndim-1] not in [1,2,3]:
          self.log.warn("only 1-3 dimensional gridding implemented")
          return 1
        else:
          self.setAttr('Eff MTX Z', visible=crds.shape[-1]==3)

        self.setAttr('Dims per Set', max = crds.ndim-1)

        if data is not None:
          if data.ndim < crds.ndim - 1:
            self.log.warn("# of dimensions of data to small")
            return 1
          else:
            for i in range(crds.ndim-1):
              if data.shape[-i-1] != crds.shape[-i-2]:
                self.log.warn("sizes of data and cords don't match")
                return 1

        if wghts is not None:
          if wghts.ndim != crds.ndim - 1:
            self.log.warn("# of dimensions of weights must match crds")
            return 1
          else:
            for i in range(crds.ndim-1):
              if crds.shape[i] != wghts.shape[i]:
                self.log.warn("sizes of weights and coords don't match")
                return 1

        # Auto Matrix calculation: extra 25% assumes "true resolution"
        if (inparam is not None):

            # header check
            if 'headerType' in inparam:
                # old ReadPhilips param output
                if inparam['headerType'] == 'BNIspiral':
                    mtx_xy = (1.25*float(inparam['spFOVXY'][0])/
                              float(inparam['spRESXY'][0]))
                    stype = int(float(inparam['spSTYPE'][0]))
                    if crds.shape[-1] == 3:
                        mtx_z  = (float(inparam['spFOVZ'][0])/
                                  float(inparam['spRESZ'][0]))
                        if stype in [2,3]:  # SDST, FLORET
                            mtx_z *= 1.25
                        self.setAttr('Eff MTX Z', val=mtx_z)

                    # Auto offset calculation.  Values reported in mm, change to # pixels
                    m_off = 0.001*float(inparam['m_offc'][0])
                    p_off = 0.001*float(inparam['p_offc'][0])
                    xoff = m_off*float(mtx_xy)/float(inparam['spFOVXY'][0])
                    yoff = p_off*float(mtx_xy)/float(inparam['spFOVXY'][0])
                    if crds.shape[-1] == 3:
                        s_off = 0.001*float(inparam['s_offc'][0])
                        zoff = s_off*float(mtx_z) /float(inparam['spFOVZ'][0])

                # new ReadPhilips param output
                elif inparam['headerType'] == 'spparams':
                    # these off-center values are already in pixels
                    stype = inparam['SPIRAL_TYPE']
                    xoff = inparam['FOV_OFFC_PIXELS'][0]
                    yoff = inparam['FOV_OFFC_PIXELS'][1]
                    zoff = inparam['FOV_OFFC_PIXELS'][2]

                    mtx_xy = (1.25*inparam['FOV_CM'][0] / inparam['RES_CM'][0])
                    if crds.shape[-1] == 3:
                        mtx_z = (inparam['FOV_CM'][2] / inparam['RES_CM'][2])
                        if stype in [2,3]:  # SDST, FLORET
                            mtx_z *= 1.25
                            zoff *= 1.25
                        self.setAttr('Eff MTX Z', val=mtx_z)
                else:
                    self.log.warn("wrong header type")
                    return 1

                self.setAttr('Eff MTX XY', val = mtx_xy)
                self.setAttr('dx (pixels)', val=xoff)
                self.setAttr('dy (pixels)', val=yoff)

                # shift half pixel when the number of slices is even with
                # distributed spirals. ZQL
                if crds.shape[-1] == 3:
                    if stype in [1,2] and int(self.getVal('Eff MTX Z'))%2 == 0:
                        zoff = zoff - 0.5
                    self.setAttr('dz (pixels)', val=zoff)
                    
            else:
                # if there is no header type, then its also the wrong type
                self.log.warn("wrong header type")
                return 1

        return 0

    def compute(self):

        import numpy as np
        import gpi_core.gridding.grid as gd

        crds = np.float32(self.getData('coords'))
        wghts = self.getData('weighting')
        data = self.getData('data')
        dps = self.getVal('Dims per Set')
        mtx_xy = (3*self.getVal('Eff MTX XY')/2) # 1.5X OVERSAMPLE
        mtx_z  = (3*self.getVal('Eff MTX Z')/2) # 1.5X OVERSAMPLE
        dx = self.getVal('dx (pixels)')
        dy = self.getVal('dy (pixels)')
        dz = self.getVal('dz (pixels)')

        # Each gridded set will have npts points gridded onto it
        # There are separate coordinates for ncsets of different gridded units (lines, planes, volumes)
        # There are an additional ndsets of data for each ncset of gridded units
        crdshape = crds[...,0].shape
        maxi = crds.ndim - 2
        npts = 1
        for i in range(dps):
          npts *= crdshape[maxi-i]
        ncsets = crds[...,0].size//npts

        # Make data or reshape it
        if data is None:
          datshape = np.array(crds[...,0].shape)
          ndsets = 1
          data = np.ones((ndsets,ncsets,npts), dtype=np.complex64)
          in_dtype = data.dtype  # make the out-type the same as the in-type
        else:
          in_dtype = data.dtype  # make the out-type the same as the in-type
          datshape = np.array(data.shape)
          ndsets = data.size//crds[...,0].size
          # Reshape data to be 3 dimensional
          # force single precision
          data = np.reshape(data.astype(np.complex64),(ndsets,ncsets,npts))

        # Reshape crds to be 3 dimensional - # sets, # pts/set, and dimensionality (1-3)
        crds = np.reshape(crds.astype(np.float32),(ncsets,npts,crds.shape[-1]))

        # Make wghts or reshape it
        if wghts is None:
          wghts = np.ones((ncsets,npts), dtype=np.float32)
        else:
          # Reshape wghts to be 2 dimensional
          wghts = np.reshape(wghts.astype(np.float32),(ncsets,npts))

        # Dimensions for array to grid on
        if crds.shape[crds.ndim-1] == 1:
          outdim = np.array([mtx_xy],dtype=np.int64)
        if crds.shape[crds.ndim-1] == 2:
          outdim = np.array([mtx_xy,mtx_xy],dtype=np.int64)
        if crds.shape[crds.ndim-1] == 3:
          outdim = np.array([mtx_xy,mtx_xy,mtx_z],dtype=np.int64)

        outdim1 = outdim
        if crds.shape[crds.ndim-1] == 3: # For Distributed Spirals
            outdim1 = np.array([mtx_z,mtx_xy,mtx_xy],dtype=np.int64)

        outshape = np.append([ndsets*ncsets], outdim1)
        out = np.zeros(outshape,  dtype = in_dtype)

        # Grid it
        for i in range(ndsets):
          for j in range(ncsets):
            outset = gd.grid(crds[j,...],data[i,j,...],wghts[j,...],outdim,dx,dy,dz)
            out[i*ncsets+j,...] = outset.astype(in_dtype)

        # RESHAPE OUTPUT
        if datshape.size == dps:
          out = out[0,...]
        else:
          outshape = np.append(datshape[0:datshape.size-dps],outdim1)
          out = np.reshape(out,outshape)

        self.setData('out', out.astype(in_dtype))

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Jun18

import copy
import gpi
import numpy as np
import gpi.GLObjects as glo

class ExternalNode(gpi.NodeAPI):
    """Reformats spin magnetization profile data generated by Spyn and Bloch
    nodes for visualization in an OpenGL viewer node.

    INPUT: MR Spin info from Bloch or Spyn module

    OUTPUTS:
    Spin Objects - graphical objects showing spins (send to GLViewer)
    PS Waveform Objects - graphical objects showing Gradient and RF waveforms (send to GLViewer)

    WIDGETS:
    Time Index - select the time point to display.  The duration per point is set by the Spyn module

    *** The following modules affect the Spin Objects List
    A/B/C Axis - select the variable to spread along the A/B/C axis
                (A/B/C are the x', y', z' axes in the GLViewer, color coded as R/G/B, respectively)
    A/B/C Axis stagger - select the distance between spins along the A/B/C axis
    Show - select a few options (can select multiple buttons)
      Axes - show the actual A/B/C axes (color coded as R/G/B)
      Text - label each axis with the variable displayed on that axis
      Mx=0 - display spin vectors without the Mx component
      My=0 - display spin vectors without the My component
      Mz=0 - display spin vectors without the Mz component
    Color Scheme - Select the color scheme for spin objects (spheres and cylinders)
      Off - white
      M3 - Mx/My/Mz fed into the R/G/B channel
      MG - color coded based on spin magnitude
      MT - color coded based on transverse magnetization
      PH - color coded basd on phase
      MTPH - same as PH, with brightness proportional to transverse magnetization
    Tip Connect - draw lines between the tips of spin vectors in the chosen direction
    Vector Radius - relative radius of cylinders used for spins
    Sphere Radius - relative radius of spheres at base of spin vectors
    Tracer History - enables drawing lines between locations of each spin-tips last n positions (sparkler effect)
    Vector History - enables drawing last n vectors

    *** The following modules affect the PS Waveform Objects List
    Waveforms - Select which waveforms to view
    Waveform Stretch - a kludgy way to adjust the aspect ratio of the PS Waveform plot (affects X-axis spacing)
    """
    ###### FOR REFERENCE ############
    # 0  1  2  3  4  5  6  7  8  9  10 11 12
    # M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt __ __
    # Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
    #################################

    def initUI(self):
        # Widgets
        self.addWidget('Slider', 'Time index')
        self.addWidget('ExclusivePushButtons', 'A Axis',
                       buttons=['T1','T2','FQ','X','Y','Z','Vx','Vy','Vz'], val=0)
        self.addWidget('DoubleSpinBox', 'A Axis stagger',min=0.,max=2.,val=0.,singlestep = 0.1,immediate=True)
        self.addWidget('ExclusivePushButtons', 'B Axis',
                       buttons=['T1','T2','FQ','X','Y','Z','Vx','Vy','Vz'], val=0)
        self.addWidget('DoubleSpinBox', 'B Axis stagger',min=0.,max=2.,val=0.,singlestep = 0.1,immediate=True)
        self.addWidget('ExclusivePushButtons', 'C Axis',
                       buttons=['T1','T2','FQ','X','Y','Z','Vx','Vy','Vz'], val=0)
        self.addWidget('DoubleSpinBox', 'C Axis stagger',min=0.,max=2.,val=0.,singlestep = 0.1,immediate=True)
        self.addWidget('Slider', 'T1 index')
        self.addWidget('Slider', 'T2 index')
        self.addWidget('Slider', 'FQ index')
        self.addWidget('Slider', 'X index')
        self.addWidget('Slider', 'Y index')
        self.addWidget('Slider', 'Z index')
        self.addWidget('Slider', 'Vx index')
        self.addWidget('Slider', 'Vy index')
        self.addWidget('Slider', 'Vz index')

        self.addWidget('NonExclusivePushButtons', 'Show',
                       buttons=['Axes','Text','Mx=0','My=0','Mz=0'],val=0)
        self.addWidget('ExclusivePushButtons', 'Color Scheme',
                       buttons=['Off','M3','MG','MT','PH','MTPH'], val=0)
        self.addWidget('ExclusivePushButtons', 'Tip Connect',
                       buttons=['OFF','T1','T2','FQ','X','Y','Z','Vx','Vy','Vz'], val=0)
        self.addWidget('Slider', 'Tip Connector Radius',val=4,max=10)
        self.addWidget('Slider', 'Vector Radius',val=4,max=10)
        self.addWidget('Slider', 'Sphere Radius')
        self.addWidget('Slider', 'Tracer History')
        self.addWidget('Slider', 'Vector History')
        self.addWidget('NonExclusivePushButtons', 'Waveforms',
                       buttons=['GX','GY','GZ','RF_r','RF_i','|RF|','RF_ph'],val=0)
        self.addWidget('Slider', 'Waveform Stretch',min=1,max=100,val=20)

        # IO Ports
        self.addInPort('M_in', 'NPYarray')
        self.addOutPort('Spin Objects', 'GLOList')
        self.addOutPort('PS Waveform Objects', 'GLOList')

        return 0

    def validate(self):
        m_in = self.getData('M_in')
        mdim = list(m_in.shape)
        options = ['Off','T1','T2','FQ','X','Y','Z','Vx','Vy','Vz']
        tipOptions = ['Off']
        aAxis = self.getVal('A Axis')
        bAxis = self.getVal('B Axis')
        cAxis = self.getVal('C Axis')

        # Set limits on index sliders
        self.setAttr('Time index',max=mdim[1]-2)
        self.setAttr('T1 index',max=mdim[2]-1)
        self.setAttr('T2 index',max=mdim[3]-1)
        self.setAttr('FQ index',max=mdim[4]-1)
        self.setAttr( 'X index',max=mdim[5]-1)
        self.setAttr( 'Y index',max=mdim[6]-1)
        self.setAttr( 'Z index',max=mdim[7]-1)
        self.setAttr('Vx index',max=mdim[8]-1)
        self.setAttr('Vy index',max=mdim[9]-1)
        self.setAttr('Vz index',max=mdim[10]-1)

        if self.getVal('Tip Connect') == 0:
          self.setAttr('Tip Connector Radius',visible = False)
        else:
          self.setAttr('Tip Connector Radius',visible = True)

        # A Axis, show options with more than 1 element
        aOptions = ['Off']
        for i in range(9):
          if mdim[i+2] > 1:
            aOptions.append(options[i+1])
        if len(aOptions) > 1:
          self.setAttr('A Axis',buttons=aOptions)
        if self.getVal('A Axis') >= len(aOptions):
          self.setAttr('A Axis',val=0)
        self.setAttr('A Axis',visible=len(aOptions) > 1)
        self.setAttr('A Axis stagger',visible=(len(aOptions) > 1) and (self.getVal('A Axis') > 0))
        if self.getVal('A Axis') > 0:
          tipOptions.append(aOptions[self.getVal('A Axis')])
         
        # B Axis, show remaining viable options
        bOptions = aOptions
        if len(bOptions) > 1:
          if self.getVal('A Axis') is not None:
            if self.getVal('A Axis') > 0: # Preserve the "Off" option
              bOptions.remove(aOptions[self.getVal('A Axis')])
        if len(bOptions) > 1:
          self.setAttr('B Axis',buttons=bOptions)
        if self.getVal('B Axis') >= len(bOptions):
          self.setAttr('B Axis',val=0)
        self.setAttr('B Axis',visible=len(bOptions) > 1)
        self.setAttr('B Axis stagger',visible=(len(bOptions) > 1) and (self.getVal('B Axis') > 0))
        if self.getVal('B Axis') > 0:
          tipOptions.append(bOptions[self.getVal('B Axis')])

        # C Axis, show remaining viable options
        cOptions = bOptions
        if len(cOptions) > 1:
          if self.getVal('B Axis') is not None:
            if self.getVal('B Axis') > 0: # Preserve the "Off" option
              cOptions.remove(bOptions[self.getVal('B Axis')])
        if len(cOptions) > 1:
          self.setAttr('C Axis',buttons=cOptions)
        if self.getVal('C Axis') >= len(cOptions):
          self.setAttr('C Axis',val=0)
        self.setAttr('C Axis',visible=len(cOptions) > 1)
        self.setAttr('C Axis stagger',visible=(len(cOptions) > 1) and (self.getVal('C Axis') > 0))
        if self.getVal('C Axis') > 0:
          tipOptions.append(cOptions[self.getVal('C Axis')])

        self.setAttr('Tip Connect',buttons=tipOptions)

        # Now Show remaining index sliders
        options.remove('Off')
        if len(cOptions) > 1:
          cOptions.remove(cOptions[self.getVal('C Axis')])
        for stuff in options:
          if cOptions.count(stuff) == 0:
            self.setAttr(stuff+' index',visible=False)
          else:
            self.setAttr(stuff+' index',visible=True)

        gwaveList = self.getVal('Waveforms')
        if gwaveList:
          self.setAttr('Waveform Stretch',visible = True)
        else:
          self.setAttr('Waveform Stretch',visible = False)

        return 0

    ############################
    def compute(self):
    ############################
    ###### FOR REFERENCE ############
    # 0  1  2  3  4  5  6  7  8  9  10 11 12
    # M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt __ __
    # Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
    #################################
        import numpy as np
        import math

        # In this module we will take m_in and pare it down
        # We will first get rid of all indices of length 1
        # Then we will slice through all dimensions that get sliced
        # and reorder the dimensions to display according to the A, B, and C axes
        # (note we will talk about A, B, C axes in the geometry viewer so we don't
        #  get confused with the X, Y, Z dimensions specified in the Spyn module)

        timeIndex = self.getVal('Time index')
        m_in = self.getData('M_in')
        mdim = list(m_in.shape)
        totalOptions = ['T1','T2','FQ','X','Y','Z','Vx','Vy','Vz']
        viableOptions = list(totalOptions)

        # Calculate the distance between points in the X/Y/Z dimensions for later use
        # This will be used to display spins in the X/Y/Z dimensions with velocity
        # Again, this is the physical X/Y/Z specified in Spyn, NOT the axes of display
        # off-xyz-scale is in units of "spin separation units per m"
        if m_in[3,1,0,0,0,0,0,0,0,0,0] != m_in[3,1,0,0,0,-1,0,0,0,0,0]:
          offxscale = (m_in.shape[5]-1)/(m_in[3,1,0,0,0,-1,0,0,0,0,0] - m_in[3,1,0,0,0,0,0,0,0,0,0])
        else:
          offxscale = 0
        if m_in[4,1,0,0,0,0,0,0,0,0,0] != m_in[4,1,0,0,0,0,-1,0,0,0,0]:
          offyscale = (m_in.shape[6]-1)/(m_in[4,1,0,0,0,0,-1,0,0,0,0] - m_in[4,1,0,0,0,0,0,0,0,0,0])
        else:
          offyscale = 0
        if m_in[5,1,0,0,0,0,0,0,0,0,0] != m_in[5,1,0,0,0,0,0,-1,0,0,0]:
          offzscale = (m_in.shape[7]-1)/(m_in[5,1,0,0,0,0,0,-1,0,0,0] - m_in[5,1,0,0,0,0,0,0,0,0,0])
        else:
          offzscale = 0
        velx = m_in[3,0,0,0,0,0,0,0,0,0,0]
        vely = m_in[4,0,0,0,0,0,0,0,0,0,0]
        velz = m_in[5,0,0,0,0,0,0,0,0,0,0]
        Tstart = m_in[6,1,0,0,0,0,0,0,0,0,0]
        xref  = m_in[3,1,0,0,0,:,0,0,0,0,0]-velx*Tstart # X position array at T=0
        yref  = m_in[4,1,0,0,0,0,:,0,0,0,0]-vely*Tstart # Y position array at T=0
        zref  = m_in[5,1,0,0,0,0,0,:,0,0,0]-velz*Tstart # Z position array at T=0

        # Remove all indices of length 1
        for i in reversed(list(range(9))):
          if mdim[i+2] == 1:
            viableOptions.pop(i)
        m_in = np.squeeze(m_in)

        # Identify axes
        aAxis = self.getVal('A Axis')
        bAxis = self.getVal('B Axis')
        cAxis = self.getVal('C Axis')
        aAxisText = "off"
        bAxisText = "off"
        cAxisText = "off"
        remainingOptions = list(viableOptions)
        axesChosen = []
        if aAxis > 0:
          axesChosen.append(remainingOptions[aAxis-1])
          aAxisText = remainingOptions[aAxis-1]
          remainingOptions.pop(aAxis-1)
        else:
          axesChosen.append('Off')
        if bAxis > 0:
          axesChosen.append(remainingOptions[bAxis-1])
          bAxisText = remainingOptions[bAxis-1]
          remainingOptions.pop(bAxis-1)
        else:
          axesChosen.append('Off')
        if cAxis > 0:
          axesChosen.append(remainingOptions[cAxis-1])
          cAxisText = remainingOptions[cAxis-1]
          remainingOptions.pop(cAxis-1)
        else:
          axesChosen.append('Off')

        # Now run through axes yet to slice
        # remaining options lists what is yet to be accounted for
        # sliceOptions gives the order of what's left in m_in, other than the 1st 2 dimensions (13,Time)
        tHist = self.getVal('Tracer History')
        vHist = self.getVal('Vector History')
        dim1wid = min(timeIndex,max(tHist, vHist))
        xi = []
        Ax = []
        xi.append(slice(None)) # Mx My Mz
        xi.append(slice(1+timeIndex-dim1wid,2+timeIndex)) # Time
        for i in range(len(viableOptions)):
          if viableOptions[i] in remainingOptions:
            sl = self.getVal(viableOptions[i]+' index')
            xi.append(sl)
          else:
            xi.append(slice(None))
            Ax.append(axesChosen.index(viableOptions[i]))
        m_in = m_in[xi]

        # OK, now m_in has between 2 and 5 dimensions
        # The first dimension has 13 indices, which includes Mx, My, and Mz
        # The second dimension has time points, just enough to do tracer and vector history
        # The next (up to) 3 dimensions are the axes along with the spins will be displayed
        # the Ax list gives which Axis the corresponding dimension belongs to
        # For example, Ax = [2,0], where m_in is 4 dimensions, means that spins are displayed 
        #             along the C axis first, the the A axis

        # Put A axis first`(in 3rd dim of m_in)
        if Ax.count(0) > 0:
          ii = Ax.index(0)
          if ii > 0:
            Ax[ii] = Ax[0]
            m_in = m_in.swapaxes(ii+2,2)
        else:
          m_in = np.expand_dims(m_in,2)
          Ax.insert(0,0)

        # Put B axis next`(in 4th dim of m_in)
        if Ax.count(1) > 0:
          ii = Ax.index(1)
          if ii > 1:
            m_in = m_in.swapaxes(ii+2,3)
        else:
          m_in = np.expand_dims(m_in,3)
          Ax.insert(1,1)

        # Put C axis last`(in 5th dim of m_in)
        # Note it should be there already, unless len(Ax) < 3
        if len(Ax) != 3:
          m_in = np.expand_dims(m_in,4)

        # Tip Connecting Logic
        tipRad = 0.01*self.getVal('Tip Connector Radius')
        tipCon = self.getVal('Tip Connect')
        if tipCon > 0 and aAxis == 0:
          tipCon += 1
        if tipCon > 1 and bAxis == 0:
          tipCon += 1

        itip = jtip = ktip = 0
        if tipCon == 1:
          itip = 1
        if tipCon == 2:
          jtip = 1
        if tipCon == 3:
          ktip = 1
        
        astag = self.getVal('A Axis stagger')
        bstag = self.getVal('B Axis stagger')
        cstag = self.getVal('C Axis stagger')
        sphereRad = 0.01*self.getVal('Sphere Radius')
        vectorRad = 0.01*self.getVal('Vector Radius')
        showList = self.getVal('Show')
        showAxes = (np.array(showList) == 0).any()
        showText = (np.array(showList) == 1).any()
        showMx   = not (np.array(showList) == 2).any()
        showMy   = not (np.array(showList) == 3).any()
        showMz   = not (np.array(showList) == 4).any()
        colorscheme = self.getVal('Color Scheme')
        amax = m_in.shape[2]
        bmax = m_in.shape[3]
        cmax = m_in.shape[4]
        aPos0 = -(amax-1)*astag*0.5
        bPos0 = -(bmax-1)*bstag*0.5
        cPos0 = -(cmax-1)*cstag*0.5
        aPos1 =  (amax-1)*astag*0.5
        bPos1 =  (bmax-1)*bstag*0.5
        cPos1 =  (cmax-1)*cstag*0.5

        #out = []
        out = glo.ObjectList()

        # Axes
        if showAxes:
          if (amax > 1):
            RGBA = (1.,0.,0.,1.)
            desc = {}
            desc = glo.Cylinder()
            desc.setSubdiv(10)
            desc.setBase(vectorRad)
            desc.setTop(vectorRad)
            desc.setRGBA(RGBA)
            desc.setP1P2((aPos0,0,0), (aPos1,0,0))
            out.append(desc)
       
          if (bmax > 1):
            RGBA = (0.,1.,0.,1.)
            desc = {}
            desc = glo.Cylinder()
            desc.setSubdiv(10)
            desc.setBase(vectorRad)
            desc.setTop(vectorRad)
            desc.setRGBA(RGBA)
            desc.setP1P2((0,bPos0,0), (0,bPos1,0))
            out.append(desc)

          if (cmax > 1):
            RGBA = (0.,0.,1.,1.)
            desc = {}
            desc = glo.Cylinder()
            desc.setSubdiv(10)
            desc.setBase(vectorRad)
            desc.setTop(vectorRad)
            desc.setRGBA(RGBA)
            desc.setP1P2((0,0,cPos0), (0,0,cPos1))
            out.append(desc)

        # Text
        if showText:
          if (amax > 1):
            RGBA = (1.,0.,0.,1.)
            desc = {}
            desc = glo.Text()
            desc.setRGBA(RGBA)
            desc.setPos((aPos1, 0, 0))
            desc.setText(aAxisText)
            out.append(desc)
          if (bmax > 1):
            RGBA = (0.,1.,0.,1.)
            desc = {}
            desc = glo.Text()
            desc.setRGBA(RGBA)
            desc.setPos((0,bPos1, 0))
            desc.setText(bAxisText)
            out.append(desc)
          if (cmax > 1):
            RGBA = (0.,0.,1.,1.)
            desc = {}
            desc = glo.Text()
            desc.setRGBA(RGBA)
            desc.setPos((0,0,cPos1))
            desc.setText(cAxisText)
            out.append(desc)

###### FOR REFERENCE ############
# 0  1  2  3  4  5  6  7  8  9  10 11 12
# M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt __ __
# Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
#################################
          
        # Now Loop through Spins
        for k in range(cmax):
          cPosbase = cPos0+k*cstag
          for j in range(bmax):
            bPosbase = bPos0+j*bstag
            for i in range(amax):
              aPosbase = aPos0+i*astag
              mx = m_in[0,dim1wid,i,j,k]
              my = m_in[1,dim1wid,i,j,k]
              mz = m_in[2,dim1wid,i,j,k]

              # This is some logic for shifting X, Y, and Z for velocities != 0
              if aAxisText == 'X':
                aPos = aPosbase + astag*(m_in[3,-1,i,j,k]-xref[i])*offxscale
              elif aAxisText == 'Y':
                aPos = aPosbase + astag*(m_in[4,-1,i,j,k]-yref[i])*offyscale
              elif aAxisText == 'Z':
                aPos = aPosbase + astag*(m_in[5,-1,i,j,k]-zref[i])*offzscale
              else:
                aPos = aPosbase
              if bAxisText == 'X':
                bPos = bPosbase + bstag*(m_in[3,-1,i,j,k]-xref[j])*offxscale
              elif bAxisText == 'Y':
                bPos = bPosbase + bstag*(m_in[4,-1,i,j,k]-yref[j])*offyscale
              elif bAxisText == 'Z':
                bPos = bPosbase + bstag*(m_in[5,-1,i,j,k]-zref[j])*offzscale
              else:
                bPos = bPosbase
              if cAxisText == 'X':
                cPos = cPosbase + cstag*(m_in[3,-1,i,j,k]-xref[k])*offxscale
              elif cAxisText == 'Y':
                cPos = cPosbase + cstag*(m_in[4,-1,i,j,k]-yref[k])*offyscale
              elif cAxisText == 'Z':
                cPos = cPosbase + cstag*(m_in[5,-1,i,j,k]-zref[k])*offzscale
              else:
                cPos = cPosbase

              # COLOR SCHEME
              # ['Off','M3','MG','MT','PH']
              brightness = 1.
              if colorscheme == 0: # Off
                red = green = blue = 1.
                clookup = 'none'
              elif colorscheme == 1: # M3
                red = abs(mx)
                green = abs(my)
                blue = abs(mz)
                clookup = 'none'
              elif colorscheme == 2: # MG
                color = 5*np.sqrt(mx*mx + my*my + mz*mz)
                clookup = 'rgb'
              elif colorscheme == 3: # MT
                color = 5*np.sqrt(mx*mx + my*my)
                clookup = 'rgb'
              elif colorscheme == 4: # PH
                color = 3.*(np.pi+np.arctan2(mx,my))/np.pi
                clookup = 'rgb'
              elif colorscheme == 5: # MT-PH
                color = 3.*(np.pi+np.arctan2(mx,my))/np.pi
                brightness = np.sqrt(mx*mx + my*my)
                clookup = 'rgb'
               
              if clookup == 'rgb':
                if color < 1:
                  red = 1.
                  green = color
                  blue = 0
                elif color < 2:
                  red = 2-color
                  green = 1.
                  blue = 0
                elif color < 3:
                  red = 0
                  green = 1.
                  blue = color-2
                elif color < 4:
                  red = 0
                  green = 4-color
                  blue = 1
                elif color < 5:
                  red = color-4
                  green = 0
                  blue = 1
                elif color <= 6:
                  red = 1.
                  green = 0
                  blue = 6-color

              red   *= brightness
              green *= brightness
              blue  *= brightness
              RGBA = (red,green,blue,1.)

              mx *= showMx
              my *= showMy
              mz *= showMz

              # VECTORS
              if vectorRad > 0.:
                desc = {}
                desc = glo.Cylinder()
                desc.setRGBA(RGBA)
                desc.setSubdiv(10)
                desc.setBase(vectorRad)
                desc.setTop(vectorRad)
                desc.setP1P2((aPos,bPos,cPos), (aPos+mx,bPos+my,cPos+mz))
                out.append(desc)

                # Vector History
                for hist in range(min(dim1wid,vHist)):
                  mx0 = m_in[0,dim1wid-1-hist,i,j,k]
                  my0 = m_in[1,dim1wid-1-hist,i,j,k]
                  mz0 = m_in[2,dim1wid-1-hist,i,j,k]
                  histRad = vectorRad*(vHist-hist)/(vHist+1.)
                  desc = {}
                  desc = glo.Cylinder()
                  desc.setRGBA(RGBA)
                  desc.setSubdiv(10)
                  desc.setBase(histRad)
                  desc.setTop(histRad)
                  desc.setP1P2((aPos,bPos,cPos), (aPos+mx0,bPos+my0,cPos+mz0))
                  out.append(desc)

                # Tracer History
                for hist in range(min(dim1wid,tHist)):
                  mx0 = m_in[0,dim1wid-hist,i,j,k]
                  my0 = m_in[1,dim1wid-hist,i,j,k]
                  mz0 = m_in[2,dim1wid-hist,i,j,k]
                  mx1 = m_in[0,dim1wid-1-hist,i,j,k]
                  my1 = m_in[1,dim1wid-1-hist,i,j,k]
                  mz1 = m_in[2,dim1wid-1-hist,i,j,k]
                  tRad0 = vectorRad*(tHist-hist)/(tHist)
                  tRad1 = vectorRad*(tHist-hist-1)/(tHist)
                  desc = {}
                  desc = glo.Cylinder()
                  desc.setRGBA(RGBA)
                  desc.setSubdiv(10)
                  desc.setBase(tRad1)
                  desc.setTop(tRad0)
                  desc.setP1P2((aPos+mx0,bPos+my0,cPos+mz0), (aPos+mx1,bPos+my1,cPos+mz1))
                  out.append(desc)

              # TIP CONNECT
              if tipCon > 0:
                if i>=itip and j>=jtip and k>=ktip:
                  mx0 = m_in[0,dim1wid,i-itip,j-jtip,k-ktip]
                  my0 = m_in[1,dim1wid,i-itip,j-jtip,k-ktip]
                  mz0 = m_in[2,dim1wid,i-itip,j-jtip,k-ktip]
                  desc = {}
                  desc = glo.Cylinder()
                  desc.setRGBA(RGBA)
                  desc.setSubdiv(10)
                  desc.setBase(tipRad)
                  desc.setTop(tipRad)
                  desc.setP1P2((aPos-itip*astag+mx0,bPos-jtip*bstag+my0,cPos-ktip*cstag+mz0), (aPos+mx,bPos+my,cPos+mz))
                  out.append(desc)

              # VECTORS
              if sphereRad > 0.:
                desc = {}
                desc = glo.Sphere()
                desc.setRadius(sphereRad)
                desc.setRGBA(RGBA)
                desc.setPos((aPos,bPos,cPos))
                desc.setSubdiv(20)
                out.append(desc)

        # Gradient Waveforms
###### FOR REFERENCE ############
# 0  1  2  3  4  5  6  7  8  9  10 11 12
# M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt __ __
# Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
#################################
        psw = glo.ObjectList()
        gwaveList = self.getVal('Waveforms')
        if gwaveList:
          wavex = 0.005*self.getVal('Waveform Stretch')
          showGx = (np.array(gwaveList)  == 0).any()
          showGy = (np.array(gwaveList)  == 1).any()
          showGz = (np.array(gwaveList)  == 2).any()
          showRfr = (np.array(gwaveList) == 3).any()
          showRfi = (np.array(gwaveList) == 4).any()
          showRfm = (np.array(gwaveList) == 5).any()
          showRfp = (np.array(gwaveList) == 6).any()
          numwaves = int(showGx) + int(showGy) + int(showGz) + int(showRfr) + int(showRfi) + int(showRfm) + int(showRfp)

          g_in = self.getData('M_in')[7:,1:,0,0,0,0,0,0,0,0,0]
          maxvals = np.amax(np.absolute(g_in), axis=1)
          gmax = np.amax(maxvals[0:3])
          rfmax = math.sqrt(maxvals[3]*maxvals[3] + maxvals[4]*maxvals[4])
          if gmax > 0:
            gnorm = 1./gmax
          else:
            gnorm = 0
          if rfmax > 0:
            rfnorm = 1./rfmax
          else:
            rfnorm = 0

          # Draw Axes
          RGBA = (1.0,1.0,1.0,1)
          x0 = 0
          x1 = g_in.shape[1] * wavex
          for i in range(numwaves):
            y = 2.5*(numwaves-1-2*i)/2
            desc = {}
            desc = glo.Cylinder()
            desc.setRGBA(RGBA)
            desc.setSubdiv(10)
            desc.setBase(0.05)
            desc.setTop(0.05)
            desc.setP1P2((x0,y,0), (x1,y,0))
            psw.append(desc)

          for i in range(timeIndex+1):
            x0 = i*wavex
            x1 = x0 - wavex
            y = 2.5*(numwaves-1)/2
            ii = max(i-1,0)
            if showGx:
              y0 = y+gnorm*g_in[0,i]
              y1 = y+gnorm*g_in[0,ii]
              RGBA = (1.0,0.5,0.5,1)
              desc = {}
              desc = glo.Cylinder()
              desc.setRGBA(RGBA)
              desc.setSubdiv(10)
              desc.setBase(0.05)
              desc.setTop(0.05)
              desc.setP1P2((x0,y0,0), (x1,y1,0))
              psw.append(desc)
              y -= 2.5
            if showGy:
              y0 = y+gnorm*g_in[1,i]
              y1 = y+gnorm*g_in[1,ii]
              RGBA = (0.5,1.0,0.5,1)
              desc = {}
              desc = glo.Cylinder()
              desc.setRGBA(RGBA)
              desc.setSubdiv(10)
              desc.setBase(0.05)
              desc.setTop(0.05)
              desc.setP1P2((x0,y0,0), (x1,y1,0))
              psw.append(desc)
              y -= 2.5
            if showGz:
              y0 = y+gnorm*g_in[2,i]
              y1 = y+gnorm*g_in[2,ii]
              RGBA = (0.5,0.5,1.0,1)
              desc = {}
              desc = glo.Cylinder()
              desc.setRGBA(RGBA)
              desc.setSubdiv(10)
              desc.setBase(0.05)
              desc.setTop(0.05)
              desc.setP1P2((x0,y0,0), (x1,y1,0))
              psw.append(desc)
              y -= 2.5
            if showRfr:
              y0 = y+rfnorm*g_in[3,i]
              y1 = y+rfnorm*g_in[3,ii]
              RGBA = (1.0,1.0,0.5,1)
              desc = {}
              desc = glo.Cylinder()
              desc.setRGBA(RGBA)
              desc.setSubdiv(10)
              desc.setBase(0.05)
              desc.setTop(0.05)
              desc.setP1P2((x0,y0,0), (x1,y1,0))
              psw.append(desc)
              y -= 2.5
            if showRfi:
              y0 = y+rfnorm*g_in[4,i]
              y1 = y+rfnorm*g_in[4,ii]
              RGBA = (1.0,0.5,1.0,1)
              desc = {}
              desc = glo.Cylinder()
              desc.setRGBA(RGBA)
              desc.setSubdiv(10)
              desc.setBase(0.05)
              desc.setTop(0.05)
              desc.setP1P2((x0,y0,0), (x1,y1,0))
              psw.append(desc)
              y -= 2.5

        if out.len() == 0: out = None
        if psw.len() == 0: psw = None
        self.setData('Spin Objects', out)
        self.setData('PS Waveform Objects', psw)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2013 Sep 17

import tempfile
import scipy.io.wavfile as spio
import numpy as np
import os
import time
from gpi import QtMultimedia

import gpi


class ExternalNode(gpi.NodeAPI):

    """Converts numpy arrays into .wav format for reproduction on the audio system,
       writes to /tmp directory, then plays .wav file.

    INPUT:
    1 dimensional real-valued array to be converted to audio.
    Data values are normalized internally.

    WIDGETS:
    Sample Rate (samp/sec) - D/A dwell time of waveform
    Loops - enter a number > 1 to play multiple times 
    Play - write waveform and play audio

    KNOWN ISSUES:
    None
"""

    def initUI(self):
        # Widgets
        self.addWidget('TextBox', 'Audio Info')
        self.addWidget('SpinBox', 'Sample Rate (samp/sec)', min=1, val=1000)
        self.addWidget('SpinBox', 'Loops', min=1, val=1)
        self.addWidget('PushButton', 'Play')

        self.addWidget(
            'SaveFileBrowser', 'File Browser', button_title='Browse',
            caption='Save File (*.wav)', filter='Wav (*.wav)')
        self.addWidget('PushButton', 'Write Now', button_title='Write Right Now', toggle=False)


        # IO Ports
        self.addInPort('wave source', 'NPYarray', ndim=1)

        self._tmpfile = tempfile.mkstemp(
            prefix='gpi_', suffix='.wav', dir='/tmp/', text=False)[1]
        self.URI = gpi.TranslateFileURI

    def validate(self):

        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

        return 0

    def compute(self):

        arr = self.getData('wave source')
        rate = self.getVal('Sample Rate (samp/sec)')
        loops = self.getVal('Loops')
        
        # make sure the wave is maximized for int16 dyn-range
        arr = (arr.astype(np.float32) / np.abs(arr)
               .max() * (pow(2, 15) - 1)).astype(np.int16)

        spio.write(self._tmpfile, rate, np.tile(arr,loops))

        try:
            s = QtMultimedia.QSound('')
            self.setAttr('Audio Info', val='Sound module is available.')
            dur = len(arr)*loops / rate
            s.play(self._tmpfile)
        except:
            self.setAttr('Audio Info', val='Sound module failed to load!\n Please visit www.github.com/gpilab/core-nodes/issues for help.')

        time.sleep(.01)
        os.remove(self._tmpfile)
        if self.getVal('Write Now') or ('File Browser' in self.widgetEvents()):

            fname = self.URI(self.getVal('File Browser'))
            if not fname.endswith('.wav'):
                fname += '.wav'

            spio.write(fname,rate,arr)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_APPLOOP



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Dallas Turley
# Date: 2013sep09

import numpy as np
import gpi

class ExternalNode(gpi.NodeAPI) :
  """Change the NPY array data type (dtype) from any type to int<n>, uint<n>, float<n> and complex<n>, where 'n' represents the numbe of bits.
  """

  def initUI(self) :
    self.addWidget('TextBox', 'info', val = '\n\n ')
    self.opbuttons = ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64', 'float16', 'float32', 'float64', 'complex64', 'complex128']
    self.addWidget('ExclusiveRadioButtons', 'OutType', buttons=self.opbuttons,val=9)

    self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
    self.addOutPort('out', 'NPYarray')

  def compute(self) :

    data = self.getData('in')
    intype = data.dtype

    choice = self.getVal('OutType')
    typ = self.opbuttons[choice]

    out = data.astype(typ)
    outtype = out.dtype

    text = "Input data type    :  " +str(intype)+"\nOutput data type :  " +str(outtype)+"\n"
    self.setAttr('info',val=text)


    self.setData('out', out)

    return 0
    
  def execType(self) :
    return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Daniel Borup
# Date: January 2019
# Notes: Adapted from Math node

import numpy as np
import gpi


class ExternalNode(gpi.NodeAPI):
    """Perform real or complex scalar operations on a per element basis.
       Three Modes of Operation:
        1) Standard - basic Arithmetic and exponential operations.
        2) Trigonometric - basic Trigonometric operations.
        3) Comparison - returns maximum, minimum, or bit mask based on
                        comparison between inputs or against Scalar.

       Operations which do not commute (e.g. divide) operate left to right, e.g.:
         output = (left port) / (right port)
    """

    def initUI(self):
        # operations
        self.button_labels = ['Left','Right']

        # Widgets
        self.choice = 0
        self.addWidget('ExclusivePushButtons', 'Choice',
                            buttons=self.button_labels, val = 0, visible = False)
        # IO Ports
        self.addInPort('inLeft', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addInPort('inRight', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''update the widgets based on the input arrays
        '''

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')

        if (set(self.portEvents()).intersection(set(['inLeft','inRight'])) or
            'Choice' in self.widgetEvents()):
            if (data1 is None and data2 is None):
                self.setAttr('Choice', visible = False)
                self.setDetailLabel('No Data')
            else:
                if data2 is None:
                    self.setAttr('Choice', visible = False)
                    self.setAttr('Choice', val=0)
                elif data1 is None:
                    self.setAttr('Choice', visible = False)
                    self.setAttr('Choice', val=1)
                else:
                    self.setAttr('Choice', visible = True)
                    if data1.shape != data2.shape:
                        self.log.warn('Array Sizes Unequal')
                self.choice = self.getVal('Choice')
                self.setDetailLabel(format(self.button_labels[self.choice]))

        return 0

    def compute(self):

        import numpy as np

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')
        self.choice = self.getVal('Choice')
        try:
            if self.choice == 0:
                self.setData('out', data1)
            else:
                self.setData('out', data2)
        except:
            return 1

        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.

# Author: Nick Zwart
# Date: 2013aug16

import gpi


class ExternalNode(gpi.NodeAPI):
    """Browse the web for data, search your local subversion repo for code
    hints, or update the GPI wiki.
    """

    def initUI(self):
        # Widgets
        self.addWidget('StringBox', 'URL', val='http://')
        self.addWidget('WebBox', 'WebBrowser', val='www.google.com')
        self.addWidget('ComboBox','GPI Bookmarks', items=['Bug Tracker','Wiki','Share','Share Uploader','Repository'])
        self.addWidget('StringBox', 'Username/Login')
        self.addWidget('StringBox', 'Password', mask=True)

    def validate(self):

        events = self.widgetEvents()
        if 'URL' in events:
            self.log.info(self.getVal('URL'))
            self.setAttr('WebBrowser', val=self.getVal('URL'), passwd=None, username=None)

        elif 'GPI Bookmarks' in events:
            mark = self.getVal('GPI Bookmarks')
            #base = 'http://mbl1:5555'
            base = 'http://cylon'
            if mark == 'Bug Tracker':
                url = base+'/trac'
            if mark == 'Wiki':
                url = base+'/wiki'
            if mark == 'Share':
                url = base+'/share'
            if mark == 'Share Uploader':
                url = base+'/ajaxplorer'
            if mark == 'Repository':
                url = base+'/svn'
            self.setAttr('WebBrowser', val=url)

        elif 'Username/Login' in events:
            self.setAttr('WebBrowser', username=self.getVal('Username/Login'))

        elif 'Password' in events:
            self.setAttr('WebBrowser', passwd=self.getAttr('Password', 'maskedval'))

        elif 'WebBrowser' in events:
            url = self.getVal('WebBrowser')
            self.setAttr('URL', val=url)

        return 0

    def execType(self):
        return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2013aug10

import copy
import numpy as np

import gpi
import gpi.GLObjects as glo


class ExternalNode(gpi.NodeAPI):
    """A basic module for generating GL object descriptions using GPI format.
    These objects can be accumulated by creating a string of GLObjects nodes, with
    the output of one node fed to the input of the next node.  The final list of GL objects
    can be viewed using the GLViewer

    INPUTS:
    GL Object List - optional input takes the output of another GLObject module, for concatenating all objects in a list
    Crds - optional input of k-space coordinates, as a numpy array, for use with the "trajectory" GL Objects function.
      The last dimension must be 3, corresponding to kx/ky/kz

    OUTPUTS - GL Object List

    WIDGETS:
    GL Objects - type of GL Object to create
      Self Evident, hook up to GLViewer to display.  Clip Plane must be combined with a GL Object to observe
    Color - changes hue of object
    Subdiv - for sphere and cylinders, specifies how many planar surfaces to approximate curve
    Pos X, Pos Y, Pos Z - X, Y, Z coordinates for center of object
    Rot X, Rot Y, Rot Y - specifies rotation of object
    Multiples - allows one to create more than one instance of object, at pseudo-random locations
    IF GL Objects = Sphere:
      Radius - Radius of Sphere
    IF GL Objects = Cylinder:
      Base, Top, Height - specifies lower and upper diameter of cylinder and its height
    IF GL Objects = Axes:
      Radius - length of axes
      Tube Radius - Diameter of axes cylinders
    IF GL Objects = Text:
      Text - string to display
    IF GL Objects = Trajectory or Random Trajectory:
      Tube Radius - diameter of cylinders used to show trajectory
    IF GL Objects = Clip Plane:
      Plane No. - Specify which of 6 clipping planes are defined
    """

    def initUI(self):
        # Widgets
        self.addWidget('ComboBox', 'GL Objects', items=['Sphere', 'Cylinder', 'Axes', 'Text', 'Trajectory', 'Random Trajectory', 'Trajectory Points', 'Clip Plane'], val='Sphere')

        self.addWidget('Slider', 'Color', min=0, max=60, val=0)
        self.addWidget('Slider', 'Subdiv', min=1, max=100, val=10)
        self.addWidget('Slider', 'Pos X', min=-100, max=100, val=0)
        self.addWidget('Slider', 'Pos Y', min=-100, max=100, val=0)
        self.addWidget('Slider', 'Pos Z', min=-100, max=100, val=0)
        self.addWidget('Slider', 'Rot X', min=0, max=360, val=0)
        self.addWidget('Slider', 'Rot Y', min=0, max=360, val=0)
        self.addWidget('Slider', 'Rot Z', min=0, max=360, val=0)
        self.addWidget('Slider', 'Multiples', min=0, max=10000, val=0)
        self.addWidget('StringBox', 'Text', val='Text')

        # Sphere
        self.addWidget('Slider', 'Radius', min=1, max=100, val=10)

        # Cylinder
        self.addWidget('Slider', 'Base', min=1, max=100, val=10)
        self.addWidget('Slider', 'Top', min=1, max=100, val=10)
        self.addWidget('Slider', 'Height', min=1, max=100, val=10)

        # Axes
        self.addWidget('Slider', 'Tube Radius', min=1, max=100, val=1)

        # ClipPlane
        self.addWidget('Slider', 'Plane No.', min=0, max=5, val=0)

        # IO Ports
        self.addInPort('GL Object List', 'GLOList', obligation=gpi.OPTIONAL)
        self.addInPort('Crds', 'NPYarray', ndim=3, vec=3, obligation=gpi.OPTIONAL)
        self.addOutPort('GL Object Descriptions', 'GLOList')

    def validate(self):

        arg = self.getVal('GL Objects')
        if arg == 'Sphere' or arg == 'Axes' or arg == 'Trajectory Points':
            self.setAttr('Radius', visible=True)
        else:
            self.setAttr('Radius', visible=False)

        if arg == 'Cylinder':
            self.setAttr('Base', visible=True)
            self.setAttr('Top', visible=True)
            self.setAttr('Height', visible=True)
        else:
            self.setAttr('Base', visible=False)
            self.setAttr('Top', visible=False)
            self.setAttr('Height', visible=False)

        if arg == 'Axes' or arg == 'Trajectory' or arg == 'Random Trajectory':
            self.setAttr('Tube Radius', visible=True)
        else:
            self.setAttr('Tube Radius', visible=False)

        if arg == 'Text':
            self.setAttr('Text', visible=True)
        else:
            self.setAttr('Text', visible=False)

        if arg == 'Clip Plane':
            self.setAttr('Plane No.', visible=True)
        else:
            self.setAttr('Plane No.', visible=False)

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''
        out = glo.ObjectList(self.getData('GL Object List'))

        arg = self.getVal('GL Objects')

        color = 0.1*self.getVal('Color')
        subdiv = self.getVal('Subdiv')
        posx = 0.1*self.getVal('Pos X')
        posy = 0.1*self.getVal('Pos Y')
        posz = 0.1*self.getVal('Pos Z')
        Rx = 1.0*self.getVal('Rot X')
        Ry = 1.0*self.getVal('Rot Y')
        Rz = 1.0*self.getVal('Rot Z')
        multiples = self.getVal('Multiples')
        text = self.getVal('Text')
        planeNo = self.getVal('Plane No.')


        # Sphere
        radius = 0.1*self.getVal('Radius')

        # Cylinder
        base = 0.1*self.getVal('Base')
        top = 0.1*self.getVal('Top')
        height = 0.1*self.getVal('Height')

        # Axes
        tuberad = 0.01*self.getVal('Tube Radius')

        if color < 1:
          red = 1.
          green = color
          blue = 0
        elif color < 2:
          red = 2-color
          green = 1.
          blue = 0
        elif color < 3:
          red = 0
          green = 1.
          blue = color-2
        elif color < 4:
          red = 0
          green = 4-color
          blue = 1
        elif color < 5:
          red = color-4
          green = 0
          blue = 1
        elif color <= 6:
          red = 1.
          green = 0
          blue = 6-color
 
        RGBA = (red, green, blue, 1.0)

        if arg == 'Sphere':
            desc = glo.Sphere()
            desc.setRadius(radius)
            desc.setRGBA(RGBA)
            desc.setPos((posx, posy, posz))
            desc.setRotXYZ((Rx, Ry, Rz))
            desc.setSubdiv(subdiv)
            if multiples:
                desc.setMultiples(np.random.rand(multiples,3)*10-5)

            out.append(desc)

        if arg == 'Clip Plane':
            desc = glo.ClipPlane()
            desc.setPos((posx, posy, posz))
            desc.setRotXYZ((Rx, Ry, Rz))
            desc.setPlaneNum(planeNo)

            out.append(desc)

        if arg == 'Cylinder':
            desc = {}
            desc = glo.Cylinder()
            desc.setRGBA(RGBA)
            desc.setPos((posx, posy, posz))
            desc.setRotXYZ((Rx, Ry, Rz))
            desc.setSubdiv(subdiv)
            desc.setBase(base)
            desc.setTop(top)
            desc.setHeight(height)
            #desc.setP1P2((0,-5,0), (0,5,0))
            if multiples:
                desc.setMultiples(np.random.rand(multiples,3)*10-5)

            out.append(desc)

        if arg == 'Axes':
            desc = {}
            desc = glo.Axes()
            desc.setRGBA(RGBA)
            desc.setPos((posx, posy, posz))
            desc.setRotXYZ((Rx, Ry, Rz))
            desc.setSubdiv(subdiv)
            desc.setTubeRadius(tuberad)
            desc.setRadius(radius)
            if multiples:
                desc.setMultiples(np.random.rand(multiples,3)*10-5)

            out.append(desc)

        if arg == 'Text':
            desc = {}
            desc = glo.Text()
            desc.setRGBA(RGBA)
            desc.setPos((posx, posy, posz))
            desc.setRotXYZ((Rx, Ry, Rz))
            desc.setSubdiv(subdiv)
            desc.setText(text)
            if multiples:
                desc.setMultiples(np.random.rand(multiples,3)*10-5)

            out.append(desc)

        if arg == 'Trajectory':

            if self.getData('Crds') is not None:
                crds = self.getData('Crds')
                crds = crds * 15  # scale to GL window

                # Use the 'multiples' option to keep a list of points instead
                # of full objects.
                for i in range(crds.shape[0]):
                    desc = glo.Cylinder()
                    desc.setMultiples(crds[i])
                    desc.setEndToEnd(True)
                    desc.setRadius(tuberad/2.0)
                    if color == 0:
                        desc.setRGBA((1,1,1,1))
                    else:
                        desc.setRGBA(RGBA)
                    desc.setSubdiv(subdiv)
                    out.append(desc)

        if arg == 'Random Trajectory':

            # Given coordinates create a new object for every segment.
            # NOTE: This is slow and memory intensive for large trajectories.

            crds = np.random.rand(10,3)*10-5

            for i in range(crds.shape[0]-1):
                desc = glo.Cylinder()
                desc.setP1P2(crds[i].tolist(), crds[i+1].tolist())
                desc.setRadius(tuberad/2.0)
                desc.setRGBA(RGBA)
                desc.setSubdiv(subdiv)
                out.append(desc)
        
        if arg == 'Trajectory Points':
                        
            if self.getData('Crds') is not None:
                crds = self.getData('Crds')
                crds = crds * 15  # scale to GL window
                            
                # Use the 'multiples' option to keep a list of points instead
                # of full objects.
                for i in range(crds.shape[0]):
                    posx = 360*crds[i,:,0]
                    posy = 360*crds[i,:,1]
                    posz = 360*crds[i,:,2]
                    desc = glo.Sphere()
                    desc.setMultiples(crds[i])
                    desc.setPos((posx, posy, posz))
                    desc.setRadius(.1*radius)
                    if color == 0:
                        desc.setRGBA((1,1,1,1))
                    else:
                        desc.setRGBA(RGBA)
                    desc.setSubdiv(subdiv)
                    out.append(desc)

        if out.len() == 0: out = None
        self.setData('GL Object Descriptions', out)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS
        #return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Daniel Borup
# Date: January 2019
# Notes: Adapted from Math node

import numpy as np
import gpi


class ExternalNode(gpi.NodeAPI):
    """Perform real or complex scalar operations on a per element basis.
       Three Modes of Operation:
        1) Standard - basic Arithmetic and exponential operations.
        2) Trigonometric - basic Trigonometric operations.
        3) Comparison - returns maximum, minimum, or bit mask based on
                        comparison between inputs or against Scalar.

       Operations which do not commute (e.g. divide) operate left to right, e.g.:
         output = (left port) / (right port)
    """

    def initUI(self):
        # operations
        self.op_labels = ['Add', 'Subtract', 'Multiply', 'Divide', 'Power',
                          'Exponential', 'LogN', 'Log10', 'Reciprocal',
                          'Conjugate', 'Magnitude', 'Sin', 'Cos', 'Tan',
                          'arcSin', 'arcCos', 'arcTan', 'arcTan2', 'Max',
                          'Min', '>', '<', '==', '!=', '>=', '<=']
        self.op = [np.add, np.subtract, np.multiply, np.divide, np.power,
                   np.exp, np.log, np.log10, np.reciprocal, np.conj, np.abs,
                   np.sin, np.cos, np.tan, np.arcsin, np.arccos, np.arctan,
                   np.arctan2, np.maximum, np.minimum, np.greater, np.less,
                   np.equal, np.not_equal, np.greater_equal, np.less_equal]
        self.trig_labels = ['Deg','Rad','Cyc']

        # Widgets
        self.inputs = 0
        self.func = 0
        self.addWidget('ExclusivePushButtons', 'Mode', buttons=[
                       'Standard', 'Trigonometric', 'Comparison'], val = 0)
        self.addWidget('ExclusivePushButtons', 'Units',
                       buttons=self.trig_labels, val = 1)
        self.addWidget('ExclusiveRadioButtons', 'Operation',
                       buttons=self.op_labels, val=0)
        self.addWidget('DoubleSpinBox', 'Scalar', val=0.0, decimals = 5)
        self.addWidget('PushButton', 'compute', toggle=True, val=1)

        # IO Ports
        self.addInPort('inLeft', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addInPort('inRight', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray')


    def validate(self):
        '''update the widgets based on the input arrays
        '''

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')

        if (set(self.portEvents()).intersection(set(['inLeft','inRight'])) or
            'Mode' in self.widgetEvents()):
            self.func = self.getVal('Mode')
            if (data1 is None and data2 is None):
                self.inputs = 0
            elif (data1 is None or data2 is None):
                self.inputs = 1
            else:
                self.inputs = 2

        if self.inputs == 2:
            if self.func == 0:
                self.op_labels = ['Add', 'Subtract', 'Multiply', 'Divide',
                                 'Power']
                self.op = [np.add, np.subtract, np.multiply, np.divide,
                           np.power]
            elif self.func == 1:
                self.op_labels = ['arcTan2']
                self.op = [np.arctan2]
            else:
                self.op_labels = ['Max', 'Min', '>', '<', '==', '!=', '>=',
                                 '<=']
                self.op = [np.maximum, np.minimum, np.greater, np.less,
                           np.equal, np.not_equal, np.greater_equal,
                           np.less_equal]
        elif self.inputs == 1:
            if self.func == 0:
                self.op_labels = ['Add', 'Subtract', 'Multiply', 'Divide',
                                 'Power', 'Exponential', 'LogN', 'Log10',
                                 'Reciprocal', 'Conjugate', 'Magnitude']
                self.op = [np.add, np.subtract, np.multiply, np.divide, np.power,
                           np.exp, np.log, np.log10, np.reciprocal, np.conj,
                           np.abs]
            elif self.func == 1:
                self.op_labels = ['Sin', 'Cos', 'Tan', 'arcSin', 'arcCos',
                                 'arcTan']
                self.op = [np.sin, np.cos, np.tan, np.arcsin, np.arccos,
                           np.arctan]
            else:
                self.op_labels = ['>', '<', '==', '!=', '>=', '<=']
                self.op = [np.greater, np.less, np.equal, np.not_equal,
                           np.greater_equal, np.less_equal]

        self.setAttr('Operation', buttons=self.op_labels)

        if self.getVal('Operation') > len(self.op):
            self.setAttr('Operation', val=0)
        operation = self.op[self.getVal('Operation')]
        if (self.inputs == 1 and
            operation in [np.add, np.subtract, np.multiply, np.divide,
                          np.power, np.greater, np.less, np.equal,
                          np.not_equal, np.greater_equal, np.less_equal]):
            self.setAttr('Scalar', visible=True)
        else:
            self.setAttr('Scalar', visible=False)

        if self.func == 1:
            self.setAttr('Units', visible = True)
        else:
            self.setAttr('Units', visible = False)

        # set the detail label
        if self.func == 1:
            funcStr = '{} ({})'.format(self.op_labels[self.getVal('Operation')],
                            self.trig_labels[self.getVal('Units')])
        else:
            funcStr = '{}'.format(self.op_labels[self.getVal('Operation')])

        if self.getAttr('Scalar', 'visible'):
            self.setDetailLabel("{}, scalar = {}".format(
                                    funcStr, self.getVal('Scalar')))
        else:
            self.setDetailLabel(funcStr)

        return 0

    def compute(self):

        import numpy as np

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')

        if data1 is not None and data1.dtype == bool:
            data1 = data1.astype(np.float)
        if data2 is not None and data2.dtype == bool:
            data2 = data2.astype(np.float)

        operation = self.op[self.getVal('Operation')]

        if self.getVal('Mode') == 1:
            units = self.getVal('Units')

        if operation in [np.sin, np.cos, np.tan]:
            if units == 0:
                if data1 is not None:
                    data1 = np.deg2rad(data1)
                elif data2 is not None:
                    data2 = np.deg2rad(data2)
            elif units == 2:
                if data1 is not None:
                    data1 = np.deg2rad(data1*360)
                elif data2 is not None:
                    data2 = np.deg2rad(data2*360)

        if self.getVal('compute'):
            if operation in [np.add, np.subtract, np.multiply, np.divide,
                             np.power, np.greater, np.less, np.equal,
                             np.not_equal, np.greater_equal, np.less_equal]:
                scalar = self.getVal('Scalar')
            else:
                scalar = None

            try:
                if self.inputs == 2:
                    out = operation(data1, data2)
                elif data1 is not None:
                    out = operation(data1, scalar)
                elif data2 is not None:
                    out = operation(data2, scalar)

                if operation in [np.arcsin, np.arccos, np.arctan, np.arctan2]:
                    if units == 0:
                        out = np.rad2deg(out)
                    elif units == 2:
                        out = np.rad2deg(out)/360

                if self.inputs != 0:
                    self.setData('out', out)
            except:
                return 1
        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS
